# coding: utf-8

"""
Shows a more advanced way of using the GPUCorrel block.

Required hardware:
  - Any camera
  - A CUDA compatible GPU
"""

import numpy as np

import crappy


if __name__ == "__main__":
  CAMERA = "Webcam"

  x = 2048
  y = 2048

  mask = np.empty((y, x), np.float32)
  r = min(x, y) / 2.
  r -= .1 * r
  r2 = r * r

  # Generating a circular weighted mask: the further from the center,
  # the lower the weight will be
  # Because why not ?
  for i in range(x):
    for j in range(y):
      mask[j, i] = max(0., 1 - ((i - x / 2) ** 2 + (j - x / 2) ** 2) /
                       (min(x, y) / 2.1) ** 2)

  # Generating your own displacement field:
  # It is simply a tuple of numpy arrays: one for the disp along X, one for Y
  # For example, let's make the fields corresponding to X translation:
  # Note: we could have used the default 'x' in this case,
  # this is just an example to show how to use custom fields
  myX = (np.ones((y, x), np.float32), np.zeros((y, x), np.float32))

  graph = crappy.blocks.Grapher(('t(s)', 'x'), ('t(s)', 'y'), ('t(s)', 'r'),
                                length=50)
  graphRes = crappy.blocks.Grapher(('t(s)', 'res'), length=50)
  graphLinDef = crappy.blocks.Grapher(('t(s)', 'Exx'), ('t(s)', 'Exy'))
  graphQuadDef = crappy.blocks.Grapher(('t(s)', 'Ux2'), ('t(s)', 'Vy2'))
  # Creating the correl block
  correl = crappy.blocks.GPUCorrel(camera=CAMERA,
                                   fields=[
                                     myX, 'y', 'r',  # Rigid body
                                     'exx', 'eyy', 'exy',   # Linear def
                                     'uxx', 'uyy', 'uxy',   # Quadratic def (x)
                                     'vxx', 'vyy', 'vxy'],  # Quadratic def (y)
                                   verbose=2,  # To print info
                                   show_diff=True,  # Display the residual
                                   # (slow!)
                                   drop=False,  # Disable data picker
                                   mask=mask,
                                   levels=4,  # Reduce the number of levels
                                   iterations=3,  # and of iteration
                                   resampling_factor=2.5,  # aggressive
                                   # resampling
                                   labels=[  # Needed to name our custom field
                                     'x', 'y', 'r', 'Exx', 'Eyy', 'Exy',
                                     'Ux2', 'Uy2', 'Uxy',
                                     'Vx2', 'Vy2', 'Vxy'],
                                   mul=3.2,  # Scalar to multiply the direction
                                   res=True)  # Ask to return the residual

  # Link compacter to main graph
  crappy.link(correl, graph)
  crappy.link(correl, graphRes)
  crappy.link(correl, graphLinDef)
  crappy.link(correl, graphQuadDef)

  crappy.start()



# coding: utf-8

import numpy as np
from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")

allowed_fields = ('x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z')


def get_field(field_string: str,
              h: int,
              w: int) -> (np.ndarray, np.ndarray):
  """Creates and returns the two fields on which the image will be projected,
  as numpy arrays.

  Args:
    field_string: The string describing the field on which to project the image
      for correlation.
    h: The height of the image.
    w: The width of the image.
  """

  if field_string == 'x':
    return (np.ones((h, w), dtype=np.float32),
            np.zeros((h, w), dtype=np.float32))

  elif field_string == 'y':
    return (np.zeros((h, w), dtype=np.float32),
            np.ones((h, w), dtype=np.float32))

  elif field_string == 'r':
    u, v = np.meshgrid(np.linspace(-w, w, w, dtype=np.float32),
                       np.linspace(-h, h, h, dtype=np.float32))
    return v * np.pi / 360, -u * np.pi / 360

  elif field_string == 'exx':
    return (np.tile(np.linspace(-w / 200, w / 200, w, dtype=np.float32),
                    (h, 1)),
            np.zeros((h, w), dtype=np.float32))

  elif field_string == 'eyy':
    return (np.zeros((h, w), dtype=np.float32),
            np.swapaxes(np.tile(np.linspace(-h / 200, h / 200, h,
                                            dtype=np.float32), (w, 1)), 1, 0))

  elif field_string == 'exy':
    return (np.swapaxes(np.tile(np.linspace(-h / 200, h / 200, h,
                                            dtype=np.float32), (w, 1)), 1, 0),
            np.zeros((h, w), dtype=np.float32))

  elif field_string == 'eyx':
    return (np.zeros((h, w), dtype=np.float32),
            np.tile(np.linspace(-w / 200, w / 200, w, dtype=np.float32),
                    (h, 1)))

  elif field_string == 'exy2':
    return (np.swapaxes(np.tile(np.linspace(-h / 200, h / 200, h,
                                            dtype=np.float32), (w, 1)), 1, 0),
            np.tile(np.linspace(-w / 200, w / 200, w, dtype=np.float32),
                    (h, 1)))

  elif field_string == 'z':
    u, v = np.meshgrid(np.linspace(-w, w, w, dtype=np.float32),
                       np.linspace(-h, h, h, dtype=np.float32))
    return u / 200, v / 200

  else:
    raise NameError(f"[DISVE Field parser] Unknown field {field_string}")


def get_res(ref: np.ndarray, img: np.ndarray, flow: np.ndarray) -> np.ndarray:
  """Calculates the difference between the original image and the one
    reconstructed from the current image and the calculated flow.

  Args:
    ref: The reference image for calculating the optical flow
    img: The current image for calculating the optical flow
    flow: The calculated optical flow
  """

  x, y = np.meshgrid(np.arange(img.shape[1]), np.arange(img.shape[0]))
  return ref - cv2.remap(img.astype(np.float32),
                         (x + flow[:, :, 0]).astype(np.float32),
                         (y + flow[:, :, 1]).astype(np.float32), 1)



# coding: utf-8

from time import time
from typing import List
from .inout import InOut
from .._global import OptionalModule

try:
  from psutil import virtual_memory
except (ModuleNotFoundError, ImportError):
  virtual_memory = OptionalModule("psutil")


class Fake_inout(InOut):
  """A class demonstrating the usage of an inout abject without requiring any
  hardware.

  It can read and/or modify the current memory usage on the computer.
  """

  def __init__(self) -> None:
    """Not much to do here."""

    super().__init__()

  def open(self) -> None:
    """Creates the buffer allowing to modify the memory usage."""

    self._buf = list()

  def set_cmd(self, *cmd: float) -> None:
    """Modifies the computer memory usage.

    If the command is lower than the current, empties the buffer.
    If the command is higher than the current, adds big lists to the buffer in
    order to use more memory.

    Args:
      *cmd (:obj:`float`): The target memory usage to set.
    """

    if not isinstance(cmd[0], float) and not isinstance(cmd[0], int):
      raise TypeError("Not the right command type for the Fake_inout !")
    if not 0 <= cmd[0] <= 100:
      raise ValueError("Command should be a percentage of memory usage !")

    if cmd[0] > virtual_memory().percent:
      self._buf.append([0] * 1024*1024)
    elif cmd[0] < virtual_memory().percent:
      try:
        # If there's nothing to delete, abort
        del self._buf[-1]
      except IndexError:
        return

  def get_data(self) -> List[float]:
    """Just returns time and the current memory usage."""

    return [time(), virtual_memory().percent]

  def close(self) -> None:
    """Deletes the buffer."""

    del self._buf



# coding: utf-8

from dataclasses import dataclass
from typing import Optional, Tuple, List


@dataclass
class Zoom:
  """This class stores the upper and lower limits of the image to display in
  the configuration window.

  It also allows updating them when the user changes the zoom ratio or drags
  the image with the mouse.
  """

  x_low: float = 0.
  x_high: float = 1.
  y_low: float = 0.
  y_high: float = 1.

  def reset(self) -> None:
    """Resets the zoom level to default (no zoom)."""

    self.x_low, self.x_high, self.y_low, self.y_high = 0, 1, 0, 1

  def update_zoom(self, x: float, y: float, ratio: float) -> None:
    """Updates the upper and lower limits of the image when the user scrolls
    with the mousewheel.

    The update is based on the zoom ratio and the position of the mouse on the
    screen.

    Args:
      x: The `x` position of the mouse on the image, as a ratio between `0`
        and `1`.
      y: The `y` position of the mouse on the image, as a ratio between `0`
        and `1`.
      ratio: The zoom ratio to apply. If it is greater than `1` we zoom in,
        otherwise we zoom out.
    """

    prev_x_low, prev_x_high = self.x_low, self.x_high
    prev_y_low, prev_y_high = self.y_low, self.y_high

    # Updating the lower x limit
    self.x_low = max(self.x_low + x * (1 - 1 / ratio), 0.)
    # Updating the upper x limit, making sure it's not out of the image
    if self.x_low + 1 / ratio * (prev_x_high - prev_x_low) > 1.:
      self.x_high = 1.
      self.x_low = 1 - 1 / ratio * (prev_x_high - prev_x_low)
    else:
      self.x_high = self.x_low + 1 / ratio * (prev_x_high - prev_x_low)

    # Updating the lower y limit
    self.y_low = max(self.y_low + y * (1 - 1 / ratio), 0.)
    # Updating the upper y limit, making sure it's not out of the image
    if self.y_low + 1 / ratio * (prev_y_high - prev_y_low) > 1.:
      self.y_high = 1.
      self.y_low = 1 - 1 / ratio * (prev_y_high - prev_y_low)
    else:
      self.y_high = self.y_low + 1 / ratio * (prev_y_high - prev_y_low)

  def update_move(self, delta_x: float, delta_y: float) -> None:
    """Updates the upper and lower limits of the image when the user moves the
    image with a left button click.

    Args:
      delta_x: The `x` displacement to apply to the image, as a ratio of the
        total image width.
      delta_y: The `y` displacement to apply to the image, as a ratio of the
        total image height.
    """

    prev_x_low, prev_x_high = self.x_low, self.x_high
    prev_y_low, prev_y_high = self.y_low, self.y_high

    # Updating the x position
    if delta_x <= 0:
      self.x_low = max(0., prev_x_low + delta_x)
      self.x_high = self.x_low + prev_x_high - prev_x_low
    else:
      self.x_high = min(1., prev_x_high + delta_x)
      self.x_low = self.x_high - prev_x_high + prev_x_low

    # Updating the y position
    if delta_y <= 0:
      self.y_low = max(0., prev_y_low + delta_y)
      self.y_high = self.y_low + prev_y_high - prev_y_low
    else:
      self.y_high = min(1., prev_y_high + delta_y)
      self.y_low = self.y_high - prev_y_high + prev_y_low


@dataclass
class Box:
  """This class represents a box to be drawn on the image of a CameraConfig
  GUI.

  It can be either the box drawn when selecting a region, or the bounding box
  of a previously detected area."""

  x_start: Optional[int] = None
  x_end: Optional[int] = None
  y_start: Optional[int] = None
  y_end: Optional[int] = None

  x_centroid: Optional[float] = None
  y_centroid: Optional[float] = None

  def no_points(self) -> bool:
    """Returns whether all four sides of the box are defined or not."""

    return any(point is None for point in (self.x_start, self.x_end,
                                           self.y_start, self.y_end))

  def reset(self) -> None:
    """Resets the sides to :obj:`None`."""

    self.x_start = None
    self.x_end = None
    self.y_start = None
    self.y_end = None

    self.x_centroid = None
    self.y_centroid = None

  def get_patch(self) -> (int, int, int, int):
    """Returns the information of the box in the patch format, for
    compatibility with other blocks."""

    return (self.y_start, self.x_start, self.y_end - self.y_start,
            self.x_end - self.x_start)

  def sorted(self) -> (int, int, int, int):
    """Returns the four sides values but sorted in the order : min x, max x,
    min y, max y."""

    if self.no_points():
      raise ValueError("Cannot sort, some values are None !")

    x_top = min(self.x_start, self.x_end)
    x_bottom = max(self.x_start, self.x_end)
    y_left = min(self.y_start, self.y_end)
    y_right = max(self.y_start, self.y_end)

    return x_top, x_bottom, y_left, y_right


@dataclass
class Spot_boxes:
  """This class stores up to four instances of :class:`Box`, defining the
  bounding boxes of the spots for video extensometry or the patches for DISVE.

  It can also instantiate the Box object by parsing a list of tuples containing
  enough information.
  """

  spot_1: Optional[Box] = None
  spot_2: Optional[Box] = None
  spot_3: Optional[Box] = None
  spot_4: Optional[Box] = None

  _index = -1

  def __getitem__(self, i: int) -> Optional[Box]:
    if i == 0:
      return self.spot_1
    elif i == 1:
      return self.spot_2
    elif i == 2:
      return self.spot_3
    elif i == 3:
      return self.spot_4
    else:
      raise IndexError

  def __setitem__(self, i: int, value: Optional[Box]) -> None:
    if i == 0:
      self.spot_1 = value
    elif i == 1:
      self.spot_2 = value
    elif i == 2:
      self.spot_3 = value
    elif i == 3:
      self.spot_4 = value
    else:
      raise IndexError

  def __iter__(self):
    self._index = -1
    return self

  def __next__(self) -> Box:
    self._index += 1
    try:
      return self[self._index]
    except IndexError:
      raise StopIteration

  def __len__(self) -> int:
    return len([spot for spot in self if spot is not None])

  def set_spots(self,
                spots: List[Tuple[int, int, int, int]]) -> None:
    """Parses a list of tuples and instantiates the corresponding Box
    objects."""

    for i, spot in enumerate(spots):
      self[i] = Box(x_start=spot[1], x_end=spot[1] + spot[3],
                    y_start=spot[0], y_end=spot[0] + spot[2])

  def empty(self) -> bool:
    """Returns :obj:`True` if all spots are :obj:`None`, else :obj:`False`."""

    return all(spot is None for spot in self)

  def reset(self) -> None:
    """Resets the boxes to :obj:`None`."""

    for i in range(4):
      self[i] = None



# coding: utf-8

"""
To test the correct resampling of the points in the Grapher.
"""

import crappy

gen = crappy.blocks.Generator([{'type': 'ramp', 'condition': None,
                                'speed': 1}], freq=500)
graph = crappy.blocks.Grapher(('t(s)', 'i_cmd'), maxpt=10, interp=True)
crappy.link(gen, graph, modifier=crappy.modifier.Integrate('cmd'))

crappy.start()



# coding: utf-8

import tkinter as tk
from typing import Optional
from copy import deepcopy
from .cameraConfigBoxes import Camera_config_with_boxes
from .cameraConfigTools import Box


class DISConfig(Camera_config_with_boxes):
  """Class similar to :ref:`Camera configuration` but also allowing to select
  the area on which the correlation will be performed.

  It is meant to be used for configuring the :ref:`Discorrel` block.
  """

  def __init__(self, camera) -> None:
    """Initializes the parent class and sets the correl box."""

    self._correl_box = Box()
    self._draw_correl_box = True

    super().__init__(camera)

  @property
  def box(self) -> Box:
    """Returns the Box object containing the region of interest."""

    return self._correl_box

  def _bind_canvas_left_click(self) -> None:
    """Binds the left mouse button click for drawing the box on which the
    correlation will be performed."""

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _start_box(self, event: tk.Event) -> None:
    """Simply saves the position of the user click, and disables the display of
    the current correl box."""

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self._select_box.x_start, \
        self._select_box.y_start = self._coord_to_pix(event.x, event.y)

    self._draw_correl_box = False

  def _stop_box(self, _: tk.Event) -> None:
    """Makes sure that the selected region is valid, sets it as the new correl
    box, and enables the display of the correl box."""

    # If it's just a regular click with no dragging, do nothing
    if self._img is None or self._select_box.no_points():
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = self._select_box.sorted()

    # If the box is flat, resetting it
    if y_left == y_right or x_top == x_bottom:
      self._select_box.reset()
      self._draw_correl_box = True
      return

    # The new correl box is just the copy of the select box
    self._correl_box = deepcopy(self._select_box)
    self._select_box.reset()

    self._draw_correl_box = True

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the select box on top
    of the displayed image."""

    # Do not draw the correl box if the user is creating the select box
    if self._draw_correl_box:
      self._draw_box(self._correl_box)
    self._draw_box(self._select_box)

    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the select box on top
    of the displayed image."""

    _, img = self._camera.get_image()

    self._cast_img(img)
    # Do not draw the correl box if the user is creating the select box
    if self._draw_correl_box:
      self._draw_box(self._correl_box)
    self._draw_box(self._select_box)
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, _: Box) -> None:
    """If the correl box is outside the image, it means that the image size has
    been modified. Simply resetting the correl box then."""

    self._correl_box.reset()



# coding: utf-8

"""
Introduction to the generator.

Shows how to use the Generator block and different paths.

No required hardware.
"""

import crappy

if __name__ == "__main__":
  # First part : a constant value (2) for 5 seconds
  path1 = {'type': 'constant', 'value': 2, 'condition': 'delay=5'}
  # Second: a sine wave of amplitude 1, freq 1Hz for 5 seconds
  path2 = {'type': 'sine', 'amplitude': 1, 'freq': 1, 'condition': 'delay=5'}
  # Third: A ramp rising a 1unit/s until the command reaches 10
  path3 = {'type': 'ramp', 'speed': 1, 'condition': 'cmd>10'}
  # Fourth : cycles of ramps: go down at 1u/s until cmd is <9
  # then go up at 2u/s for 1s. Repeat 5 times
  path4 = {'type': 'cyclic_ramp', 'speed1': -1, 'condition1': 'cmd<9',
           'speed2': 2, 'condition2': 'delay=1', 'cycles': 5}

  # The generator: takes the list of all the paths to be generated
  # cmd_label specifies the name to give the signal
  # freq : the target of points/s
  # spam : Send the value even if nothing changed
  #   (so the graph updates continuously)
  # verbose : add some information in the terminal
  gen = crappy.blocks.Generator([path1, path2, path3, path4],
                                cmd_label='cmd', freq=50, spam=True,
                                verbose=True)

  # The graph : we will plot cmd over time
  graph = crappy.blocks.Grapher(('t(s)', 'cmd'))

  # Do not forget to link them or the graph will have nothing to plot !
  crappy.link(gen, graph)

  # Let's start the program
  crappy.start()



# coding: utf-8

from time import time
from typing import Tuple, Optional
from numpy import ndarray
from .camera import Camera
from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class Webcam(Camera):
  """A basic class for reading images from a USB camera (including webcams).

  It relies on the OpenCv library. Note that it was purposely kept extremely
  simple as it is mainly used as a demo. See :ref:`Camera OpenCV` and
  :ref:`Camera GStreamer` for classes giving a finer control over the device.
  """

  def __init__(self) -> None:
    """Sets variables and adds the channels setting."""

    super().__init__()
    self.name = "webcam"
    self._cap = None
    self.add_choice_setting(name="channels",
                            choices=('1', '3'),
                            default='1')

  def open(self, device_num: Optional[int] = 0, **kwargs) -> None:
    """Opens the video stream and sets any user-specified settings.

    Args:
      device_num (:obj:`int`, optional): The number of the device to open.
      **kwargs: Any additional setting to set before opening the graphical
        interface.
    """

    # Opening the videocapture device
    self._cap = cv2.VideoCapture(device_num)

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Grabs a frame from the videocapture object and returns it along with a
    timestamp."""

    # Grabbing the frame and the timestamp
    t = time()
    ret, frame = self._cap.read()

    # Checking the integrity of the frame
    if not ret:
      raise IOError("Error reading the camera")

    # Returning the image in the right format, and its timestamp
    if self.channels == '1':
      return t, cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    else:
      return t, frame

  def close(self) -> None:
    """Releases the videocapture object."""

    if self._cap is not None:
      self._cap.release()



# coding: utf-8
# Blocks are classes, running indefinitely in a single process.
# Some of them are already implemented (see the reference manual),
# but you can also implement your own.

from .autoDrive import AutoDrive
from .block import Block
from .camera import Camera
from .client_server import Client_server
from .dashboard import Dashboard
from .discorrel import DISCorrel
from .displayer import Displayer
from .disve import DISVE
from .drawing import Drawing
from .fake_machine import Fake_machine
from .generator import Generator
from .gpucorrel import GPUCorrel
from .gpuve import GPUVE
from .grapher import Grapher
from .gui import GUI
from .hdf_recorder import Hdf_recorder, Hdf_saver
from .ioblock import IOBlock
from .machine import Machine
from .mean import Mean_block
from .multiplex import Multiplex
from .pid import PID
from .reader import Reader
from .recorder import Recorder, Saver
from .sink import Sink
from .ucontroller import UController
from .videoExtenso import Video_extenso



# coding: utf-8

from typing import Callable, Union, Optional, Tuple, Dict, Any
from pathlib import Path
import numpy as np
from time import time, strftime, gmtime
from re import fullmatch
from types import MethodType

from .block import Block
from .displayer import Displayer
from ..camera import camera_list, Camera as BaseCam
from ..tool import Camera_config, Box
from .._global import OptionalModule

try:
  import SimpleITK as Sitk
except (ModuleNotFoundError, ImportError):
  Sitk = OptionalModule("SimpleITK")

try:
  import PIL
except (ModuleNotFoundError, ImportError):
  PIL = OptionalModule("Pillow")

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")


class Camera(Block):
  """This block simply acquires images from a camera.

  It can then save the images, and / or display them. The image acquisition can
  be triggered via incoming links. Optionally, a configuration window can be
  displayed for interactively tuning the camera settings before the test
  starts.

  This class also serves as a base class for other blocks that perform image
  processing on the acquired frames.
  """

  cam_count = dict()

  def __init__(self,
               camera: str,
               transform: Optional[Callable[[np.ndarray], np.ndarray]] = None,
               config: bool = True,
               display_images: bool = False,
               displayer_backend: Optional[str] = None,
               displayer_framerate: float = 5,
               software_trig_label: Optional[str] = None,
               verbose: bool = False,
               freq: float = 200,
               save_images: bool = False,
               img_name: str = "{self._n_loops:6d}_{t-self.t0:.6f}.tiff",
               save_folder: Optional[Union[str, Path]] = None,
               save_period: int = 1,
               save_backend: Optional[str] = None,
               image_generator: Optional[Callable[[float, float],
                                                  np.ndarray]] = None,
               **kwargs) -> None:
    """Sets the args and initializes the parent class.

    Args:
      camera: The name of the camera to control. See :ref:`Cameras` for an
        exhaustive list of available cameras.
      transform: A function taking an image as an argument and returning a
        transformed image. The original image is discarded and only the
        transformed one is kept for processing, display and saving.
      config: If :obj:`True`, a config window is shown before the test starts
        for interactively tuning the camera settings. It also allows selecting
        the spots to track.
      display_images: If :obj:`True`, a window displays the acquired images
        in low resolution during the test. This display is mainly intended for
        debugging and visual follow-up, but not for displaying high-quality
        images.
      displayer_backend: If ``display_images`` is :obj:`True`, the backend to
        use for the display window. Should be one of :
        ::

          'cv2', 'mpl'

        If not given, OpenCV will be used if available.
      displayer_framerate: If ``display_images`` is :obj:`True`, sets the
        maximum framerate for updating the display window. This setting allows
        limiting the resources used by the displayer. Note that the actual
        achieved framerate might differ, this is just the maximum limit.
      software_trig_label: If given, the block will only acquire images when
        receiving data on this label. The received data can be anything, even
        empty. This label will thus de facto act as a software trigger for the
        camera.
      verbose: If :obj:`True`, the achieved framerate will be displayed in the
        console during the test.
      freq: If given, the block will try to loop at this frequency. If it is
        lower than the framerate of the camera, frames will be dropped. This
        argument can be used for limiting the achieved framerate when the
        camera doesn't support framerate control.
      save_images: If :obj:`True`, the acquired images are saved on the
        computer during the test. Note that saving images uses CPU, so the
        achieved performance might drop when this feature is in use.
      img_name: If ``save_images`` is :obj:`True`, the template for naming the
        recorded images. It is evaluated as an `f-string`, and must contain the
        file extension at the end. For building the `f-string`, the
        ``self._n_loops`` attribute holds the loop number, and ``t-self.t0``
        holds the current timestamp.
      save_folder: If ``save_images`` is :obj:`True`, the directory to save
        images to. If it doesn't exist, it will be created. If not given, the
        images are saved in a folder named `Crappy_images` and created next to
        the file being run.
      save_period: If ``save_images`` is :obj:`True`, only one every this
        number of images will be saved.
      save_backend: The backend to use for saving the images. Should be one
        of :
        ::

          'sitk', 'pil', 'cv2'

        If not specified, SimpleITK will be used if available, then OpenCV as a
        second choice, and finally Pillow if none of the others was available.
      image_generator: A function taking two floats as arguments, and returning
        an image. It is only used for demonstration without camera in the
        examples, and isn't meant to be used in an actual test.
      **kwargs: Any additional argument to pass to the camera.
    """

    super().__init__()

    self.verbose = verbose
    self.freq = freq
    self.niceness = -10

    # Checking if the requested camera exists in Crappy
    if image_generator is None:
      if camera.capitalize() not in camera_list:
        raise ValueError(f"No camera named {camera.capitalize()} found in the "
                         f"list of available cameras !")
      self._camera_name = camera.capitalize()
    else:
      self._camera_name = 'Image Generator'

    # Counting the number of instantiated cameras for each type
    if self._camera_name not in Camera.cam_count:
      Camera.cam_count[self._camera_name] = 1
    else:
      Camera.cam_count[self._camera_name] += 1

    # Trying the different possible backends and checking if the given one
    # is correct
    if save_images:
      if save_backend is None:
        if not isinstance(Sitk, OptionalModule):
          self._save_backend = 'sitk'
        elif not isinstance(cv2, OptionalModule):
          self._save_backend = 'cv2'
        elif not isinstance(PIL, OptionalModule):
          self._save_backend = 'pil'
        else:
          raise ModuleNotFoundError("Neither SimpleITK, opencv-python nor "
                                    "Pillow could be imported, no backend "
                                    "found for saving the images")
      elif save_backend in ('sitk', 'pil', 'cv2'):
        self._save_backend = save_backend
      else:
        raise ValueError("The save_backend argument should be either 'sitk', "
                         "'pil' or 'cv2' !")
    else:
      self._save_backend = None

    # Checking that the given image name is valid
    if save_images:
      if fullmatch(r'.*\{.+}.*', img_name) is None:
        raise ValueError("img_name cannot be evaluated as a regular "
                         "expression !")
      elif fullmatch(r'.+\..+', img_name) is None:
        raise ValueError("No extension given in the img_name argument !")
      self._img_name = img_name
    else:
      self._img_name = None

    # Setting a default save folder if not given
    if save_images:
      if save_folder is None:
        self._save_folder = Path.cwd() / 'Crappy_images'
      else:
        self._save_folder = Path(save_folder)
    else:
      self._save_folder = None

    self._save_period = int(save_period)

    # Instantiating the displayer window if requested
    if display_images:
      self._displayer = Displayer(f"Displayer {camera} "
                                  f"{Camera.cam_count[self._camera_name]}",
                                  displayer_framerate,
                                  displayer_backend)
    else:
      self._displayer = None

    # Setting the other attributes
    self._save_images = save_images
    self._trig_label = software_trig_label
    self._config_cam = config
    self._transform = transform
    self._image_generator = image_generator
    self._camera_kwargs = kwargs

    self._n_loops = 0

  def prepare(self) -> None:
    """Preparing the save folder, opening the camera and displaying the
    configuration GUI."""

    # Creating the folder for saving the images if it doesn't exist
    if self._save_folder is not None:
      if not self._save_folder.exists():
        Path.mkdir(self._save_folder, exist_ok=True, parents=True)

    # Case when the images are generated and not acquired
    if self._image_generator is not None:
      self._camera = BaseCam()
      self._camera.add_scale_setting('Exx', 0., 100., None, None, 0.)
      self._camera.add_scale_setting('Eyy', 0., 100., None, None, 0.)
      self._camera.set_all()

      def get_image(self_) -> Tuple[float, np.ndarray]:
        return time(), self._image_generator(self_.Exx, self_.Eyy)

      self._camera.get_image = MethodType(get_image, self._camera)

    # Case when an actual camera object is responsible for acquiring the images
    else:
      self._camera = camera_list[self._camera_name]()
      self._camera.open(**self._camera_kwargs)

    if self._config_cam:
      config = Camera_config(self._camera)
      config.main()

    if self._displayer is not None:
      self._displayer.prepare()

  def loop(self) -> None:
    """Receives the incoming data, acquires an image, displays it, saves it,
    and finally processes it if needed."""

    data = self.recv_all_last()

    # Waiting for the trig label if it was given
    if self._trig_label is not None and self._trig_label not in data:
      return

    # Updating the image generator if there's one
    if self._image_generator is not None:
      if 'Exx(%)' in data:
        self._camera.Exx = data['Exx(%)']
      if 'Eyy(%)' in data:
        self._camera.Eyy = data['Eyy(%)']

    # Actually getting the image from the camera object
    ret = self._camera.get_image()
    if ret is None:
      return
    metadata, img = ret

    # Building the metadata if it was not provided
    if isinstance(metadata, float):
      metadata = {'t(s)': metadata,
                  'DateTimeOriginal': strftime("%Y:%m:%d %H:%M:%S",
                                               gmtime(metadata)),
                  'SubsecTimeOriginal': f'{metadata % 1:.6f}',
                  'ImageUniqueID': self._n_loops}
    metadata['t(s)'] -= self.t0

    self._n_loops += 1

    # Applying the transform function
    if self._transform is not None:
      img = self._transform(img)

    # Updating the displayer
    if self._displayer is not None:
      self._displayer.update(img)

    # Saving the image
    if self._save_images and not self._n_loops % self._save_period:
      path = str(self._save_folder / eval('f"{}"'.format(self._img_name)))
      self._save(img, path)

    # Performing the additional actions for subclasses
    self._additional_loop(metadata, img)

  def finish(self) -> None:
    """Closes the camera and the displayer."""

    if self._image_generator is None:
      self._camera.close()

    if self._displayer is not None:
      self._displayer.finish()

  def _save(self, img: np.ndarray, path: str) -> None:
    """Simply saves the given image to the given path using the selected
    backend."""

    if self._save_backend == 'sitk':
      Sitk.WriteImage(Sitk.GetImageFromArray(img), path)

    elif self._save_backend == 'cv2':
      cv2.imwrite(path, img)

    elif self._save_backend == 'pil':
      PIL.Image.fromarray(img).save(path)

  def _additional_loop(self, meta: Dict[str, Any], img: np.ndarray) -> None:
    """Additional action to perform in the loop, used by subclasses of the
    Camera block."""

    ...

  @staticmethod
  def _draw_box(img: np.ndarray, box: Box) -> None:
    """Draws a box on top of an image."""

    if box.no_points():
      return

    x_top, x_bottom, y_left, y_right = box.sorted()

    for line in ((box.y_start, slice(x_top, x_bottom)),
                 (box.y_end, slice(x_top, x_bottom)),
                 (slice(y_left, y_right), x_top),
                 (slice(y_left, y_right), x_bottom),
                 (box.y_start + 1, slice(x_top, x_bottom)),
                 (box.y_end - 1, slice(x_top, x_bottom)),
                 (slice(y_left, y_right), x_top + 1),
                 (slice(y_left, y_right), x_bottom - 1)
                 ):
      img[line] = 255 * int(np.mean(img[line]) < 128)



# coding: utf-8

"""
Just like ``read.py``, but with both output and input.

This program will send a sine wave on the first output and read the first
input. The type of board can be chosen when starting the file.

Required hardware:
  - Any DAQ board
"""

import crappy

if __name__ == "__main__":
  for i, c in enumerate(crappy.inout.inout_dict):
    print(i, c)
  name = list(crappy.inout.inout_dict.keys())[int(input(
      "What board do you want to use ?> "))]

  sg = crappy.blocks.Generator([{'type': 'sine', 'freq': .5, 'amplitude': 1,
                                 'offset': .5, 'condition': 'delay=1000'}],
                               cmd_label='cmd')

  io = crappy.blocks.IOBlock(name, labels=['t(s)', 'chan0'],
                             cmd_labels=['cmd'], out_channels=0, verbose=True)
  crappy.link(sg, io)

  g = crappy.blocks.Grapher(('t(s)', 'chan0'))

  crappy.link(io, g)

  crappy.start()



# coding: utf-8

from time import time
import numpy as np
from typing import Callable, Dict, Optional

from .block import Block


def plastic(v: float, yield_strain: float = .005, rate: float = .02) -> float:
  """A basic plastic law given as an example."""

  if v > yield_strain:
    return rate * ((((v - yield_strain) / rate) ** 2 + 1) ** .5 - 1)
  return 0


class Fake_machine(Block):
  """This block simulates the behavior of a tensile test machine.

  It should be used to simulate tensile teste, not compression tests. By
  default, it assumes a plastic behavior of the material. The main mechanical
  parameters of the material are tunable.

  This block is meant to be driven like the :ref:`Machine` block. However, its
  outputs are different and are : ``t(s), F(N), x(mm), Exx(%), Eyy(%)``.
  """

  def __init__(self,
               k: float = 8.4E6,
               l0: float = 200,
               max_strain: float = 1.51,
               sigma: Optional[Dict[str, float]] = None,
               nu: float = 0.3,
               plastic_law: Callable[[float], float] = plastic,
               max_speed: float = 5,
               mode: str = 'speed',
               cmd_label: str = 'cmd',
               freq: float = 100,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      k: The rigidity of the material, in N, so that ``force = k x strain``.
      l0: The initial length of the fake sample to test, in mm.
      max_strain: The maximum strain the material can withstand before
        breaking.
      mode: Whether the command sent to the fake machine is a speed or a
        position command. Can be ``'speed'`` or ``'position'``.
      plastic_law: A callable taking the maximum reached strain and returning
        the proportion of the current strain caused by plastic deformation.
      sigma: A :obj:`dict` containing for each label the standard deviation for
        adding noise to the signal. Can be given for part or all of the labels.
        The deviation should be given not normalized, in the same unit as the
        label to which it applies.
      nu: Poisson's ratio of the material.
      cmd_label: The label carrying the command of the fake machine.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.
    """

    super().__init__()
    self.freq = freq
    self.verbose = verbose

    # Setting the mechanical parameters of the material
    self._k = k
    self._l0 = l0
    self._max_strain = max_strain / 100
    self._nu = nu
    self._max_speed = max_speed
    self._sigma = {'F(N)': 50, 'x(mm)': 2e-3,
                   'Exx(%)': 1e-3, 'Eyy(%)': 1e-3} if sigma is None else sigma
    self._plastic_law = plastic_law

    self._mode = mode
    self._cmd_label = cmd_label

    # Creating the mechanical variables
    self._current_pos = 0
    self._last_t = None
    self._plastic_elongation = 0
    self._max_recorded_strain = 0

  def prepare(self) -> None:
    """Sends a value so that the camera gets an image during prepare for the
    fake test examples. To be removed at some point."""

    # Todo: To be removed when compute_strain_img is replaced with a getter
    self.t0 = time()
    self._send_values()

  def begin(self) -> None:
    """Sends a first value that should be 0."""

    self._last_t = self.t0
    self._send_values()

  def loop(self) -> None:
    """Receives the latest command value, calculates the new speed and position
    from it, checks whether the sample broke and what the plastic elongation
    is, and finally returns the data"""

    # Getting the latest command
    cmd = self.get_last()[self._cmd_label]
    t = time()
    delta_t = t - self._last_t
    self._last_t = t

    # Calculating the speed based on the command and the mode
    if self._mode == 'speed':
      speed = np.sign(cmd) * np.min((self._max_speed, np.abs(cmd)))
    elif self._mode == 'position':
      speed = np.sign(cmd - self._current_pos) * np.min(
          (self._max_speed, np.abs(cmd - self._current_pos) / delta_t))
    else:
      raise ValueError(f'Invalid mode : {self._mode} !')

    # Updating the current position
    self._current_pos += speed * delta_t

    # If the max strain is reached, consider that the sample broke
    if self._current_pos / self._l0 > self._max_strain:
      self._k = 0

    # Compute the plastic elongation separately
    if self._current_pos / self._l0 > self._max_recorded_strain:
      self._max_recorded_strain = self._current_pos / self._l0
      self._plastic_elongation = self._plastic_law(
        self._max_recorded_strain) * self._l0

    # Finally, sending the values
    self._send_values()

  def _add_noise(self, to_send: Dict[str, float]) -> Dict[str, float]:
    """Adds noise to the data to be sent, according to the sigma values
    provided by the user. Then returns the noised data."""

    for label, value in to_send.items():
      if label in self._sigma:
        to_send[label] = np.random.normal(value, self._sigma[label])

    return to_send

  def _send_values(self) -> None:
    """Gathers all the information to be sent, adds noise and send it."""

    to_send = {'t(s)': time() - self.t0,
               'F(N)': (self._current_pos -
                        self._plastic_elongation) / self._l0 * self._k,
               'x(mm)': self._current_pos,
               'Exx(%)': self._current_pos * 100 / self._l0,
               'Eyy(%)': -self._nu * self._current_pos * 100 / self._l0}

    self.send(self._add_noise(to_send))



# coding: utf-8

from time import time
import numpy as np
from typing import List, Optional, Dict, Any
from dataclasses import dataclass, field
from re import fullmatch
from collections import defaultdict
from itertools import chain

from .inout import InOut
from .._global import OptionalModule
try:
  import nidaqmx
  from nidaqmx import stream_readers, stream_writers
except (ModuleNotFoundError, ImportError):
  nidaqmx = OptionalModule("nidaqmx")
  stream_readers = stream_writers = nidaqmx

thcp_map = {"B": 10047,
            "E": 10055,
            "J": 10072,
            "K": 10073,
            "N": 10077,
            "R": 10082,
            "S": 10085,
            "T": 10086}

unit_map = {"C": 10143,
            "F": 10144,
            "R": 10145,
            "K": 10325}


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a NI DAQmx
  channel can have."""

  meas_type: str = 'voltage'
  name: Optional[str] = None
  kwargs: Dict[str, Any] = field(default_factory=dict)

  def update(self, dic_in: Dict[str, Any]) -> None:
    """Updates the channel keys based on the user input."""

    for key, val in dic_in.items():
      # The 'name' and 'type' keys are handled separately
      if key == 'name':
        self.name = val
      elif key == 'type':
        self.meas_type = val

      # All the other keys are put together in the kwargs attribute
      else:
        self.kwargs.update({key: val})


class Nidaqmx(InOut):
  """This class can communicate with NI DAQmx devices using the :mod:`pydaqmx`
  module.

  It can read single data points from digital and analog channels, read streams
  of data from analog channels, and set the voltage of analog and digital
  output channels. For analog input channels, several types of acquisition can
  be performed, like voltage, resistance, current, etc.
  """

  def __init__(self,
               channels: List[Dict[str, Any]],
               sample_rate: float = 100,
               n_samples: Optional[int] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      channels: A :obj:`list` containing :obj:`dict` holding information on the
        channels to read data from or write data to. See below for the
        mandatory and optional keys for the dicts. Note that in streamer mode,
        the digital input channels are not available for reading. Also, only
        one type of analog input channel at a time can be read in streamer
        mode, with no restriction on the number of channels of this type.
      sample_rate: The target sample rate for data acquisition in streamer
        mode, given as a :obj:`float`. Default is `100` SPS.
      n_samples: The number of samples to acquire per chunk of data in streamer
        mode. Default is 20% of ``sample_rate``.


    Note:
      - ``channels`` keys:

        - name: The name of the channel to interface with, given with the
          following syntax :
          ::

            'DevX/[a/d][i/o]Y'

          With `X` the index of the device, and `Y` the line on which the
          channel is. `d` stands for digital, `a` for analog, `i` for input and
          `o` for output. For digital channels, `DevX/d[i/o]Y` is internally
          converted to `DevX/port<Y // 8>/line<Y % 8>`. Example of a valid
          name : `Dev1/ao3`.

        - type: The type of data to read, for analog input channels. This field
          can take many different values, refer to the documentation of the
          :mod:`nidaqmx` for more details. This field is internally used for
          calling the method : :meth:`nidaqmx.task.add_ai_[type]_chan`. The
          default for this field is `'voltage'`, possible values include
          `'thrmcpl'`, `'bridge'`, `'current'` and `'resistance'`.

        - All the other keys will be given as kwargs to the
          :meth:`nidaqmx.task.add_ai_[type]_chan` method for analog input
          channels, to the :meth:`nidaqmx.task.add_ao_voltage_chan` for analog
          output channels, to :meth:`nidaqmx.task.add_do_chan` for digital
          output channels, and to :meth:`nidaqmx.task.add_di_chan` for digital
          input channels. Refer to :mod:`nidaqmx` documentation for the
          possible arguments and values. Note that for the `'thrmcpl'` analog
          input channel type, the `'thermocouple_type'` argument must be given
          as a letter, same for the `'units'` argument. They will be parsed
          internally. Also note that for the analog output channels and the
          analog input channels of type `'voltage'`, the `'min_val'` and
          `'max_val'` arguments are internally set by default to `0` and `5`.
    """

    super().__init__()

    # Setting the number of samples per acquisition for streamer mode
    if n_samples is None:
      self._n_samples = max(1, int(sample_rate / 5))
    else:
      self._n_samples = n_samples

    self._sample_rate = sample_rate

    # These attributes will be set later
    self._task_ao = None
    self._stream_ao = None
    self._task_di = None
    self._stream_di = None
    self._task_do = None
    self._stream_do = None
    self._stream_started = False
    # For analog inputs a dict is needed as each type is handled separately
    self._tasks_ai = dict()
    self._stream_ai = dict()

    self._digital_in = list()
    self._digital_out = list()
    self._analog_out = list()
    # Here as well a dict is needed for handling each analog input type
    self._analog_in = defaultdict(list)

    for channel in channels:

      # Making sure each channel has a 'name' attribute
      if 'name' not in channel:
        raise AttributeError("[NI DAQmx] The given channels must contain "
                             "the 'name' key !")

      # Parsing the channel name to retrieve info from it
      match = fullmatch(r'(.+)/(.+)(\d+)', channel['name'])
      if match is not None:
        dev, type_, num = match.groups()
        num = int(num)
      else:
        raise AttributeError(f"[NI DAQmx] Invalid format for the channel name "
                             f": {channel['name']} !\nIt should be "
                             f"'Dev<dev num>/[a/d][i/o]<chan num>'")

      # Creating a _Channel object holding the information on the channel
      chan = _Channel()
      chan.update(channel)

      # Saving the channel to the right place and performing specific actions
      if type_ == 'ai':
        self._analog_in[chan.meas_type].append(chan)
      elif type_ == 'ao':
        self._analog_out.append(chan)
      elif type_ == 'di':
        chan.name = f"{dev}/port{num // 8}/line{num % 8}"
        self._digital_in.append(chan)
      elif type_ == 'do':
        chan.name = f"{dev}/port{num // 8}/line{num % 8}"
        self._digital_out.append(chan)
      else:
        raise ValueError(f"[NI DAQmx] Wrong channel type : {type_} !\nIt "
                         f"should be either 'ai', 'ao', 'di', or 'do'.")

  def open(self) -> None:
    """Creates tasks and streams for analog output, digital input, digital
    output, and each type of analog input channels."""

    # Creating one task for each type of analog input channel
    self._tasks_ai = {type_: nidaqmx.Task() for type_ in self._analog_in}

    # Iterating over all the analog input channels
    for type_, channels in self._analog_in.items():
      for chan in channels:

        # Setting the min and max voltage for the voltage analog input channels
        if type_ == 'voltage':
          chan.kwargs['max_val'] = chan.kwargs.get('max_val', 5)
          chan.kwargs['min_val'] = chan.kwargs.get('min_val', 0)

        # Parsing the thermocouple related arguments
        if 'thermocouple_type' in chan.kwargs:
          chan.kwargs['thermocouple_type'] = thcp_map[
            chan.kwargs['thermocouple_type']]
          # Included in the if as 'units' is an arg for other types of channels
          if 'units' in chan.kwargs:
            chan.kwargs['units'] = thcp_map[chan.kwargs['units']]

        # Adding the channel to the task with the given kwargs
        try:
          func = getattr(self._tasks_ai[type_].ai_channels,
                         f'add_ai_{type_}_chan')
          func(chan.name, **chan.kwargs)
        except AttributeError:
          raise ValueError(f"[NI DAQmx] Invalid channel type : {type_}")

    # Opening a stream for each analog input task
    self._stream_ai = {
      type_: stream_readers.AnalogMultiChannelReader(task.in_stream)
      for type_, task in self._tasks_ai.items()}

    # Creating a task and a stream for all the analog output channels
    if self._analog_out:
      self._task_ao = nidaqmx.Task()
      self._stream_ao = stream_writers.AnalogMultiChannelWriter(
        self._task_ao.out_stream, auto_start=True)

      # Setting the min and max voltage for all the analog output channels
      for chan in self._analog_out:
        chan.kwargs['max_val'] = chan.kwargs.get('max_val', 5)
        chan.kwargs['min_val'] = chan.kwargs.get('min_val', 0)
        self._task_ao.ao_channels.add_ao_voltage_chan(chan.name, **chan.kwargs)

    # Creating a task and a stream for all the digital input channels
    if self._digital_in:
      self._task_di = nidaqmx.Task()

      for chan in self._digital_in:
        self._task_di.di_channels.add_di_chan(chan.name, **chan.kwargs)

      self._stream_di = stream_readers.DigitalMultiChannelReader(
        self._task_di.in_stream)

    # Creating a task and a stream for all the digital output channels
    if self._digital_out:
      self._task_do = nidaqmx.Task()

      for chan in self._digital_out:
        self._task_do.do_channels.add_do_chan(chan.name, **chan.kwargs)

      self._stream_do = stream_writers.DigitalMultiChannelWriter(
        self._task_do.out_stream)

  def start_stream(self) -> None:
    """Starts the streaming task for analog input channels.

    Data can be acquired via streaming for multiple channels, but only for one
    type of channel.
    """

    # Making sure there's only one type of channel to read data from
    if len(self._tasks_ai) > 1:
      raise IOError("[NI DAQmx] Stream mode can only open one type of "
                    "channel !")
    elif len(self._tasks_ai) < 1:
      raise IOError("[NI DAQmx] There's no analog in channel to read data "
                    "from !")

    # Starting the streaming task, there should be only one
    for task in self._tasks_ai.values():
      task.timing.cfg_samp_clk_timing(
        self._sample_rate,
        sample_mode=nidaqmx.constants.AcquisitionType.CONTINUOUS)

    self._stream_started = True

  def get_data(self) -> List[float]:
    """Reads data from the analog and digital input channels, and returns it
    along with a timestamp.

    Data from the analog channels is read first, and then data from the digital
    channels. Data is returned in the same order as it was acquired.
    """

    ret = [time()]

    # Reading the analog channels
    if self._analog_in:
      data = np.empty(len(list(chain(*self._analog_in.values()))))
      i = 0
      for type_, stream in self._stream_ai.items():
        stream.read_one_sample(data[i:i + len(self._analog_in[type_])])
        i += len(self._analog_in[type_])

      ret.extend(list(data))

    # Reading the digital channels
    if self._digital_in:
      data = np.empty((len(self._digital_in), 1), dtype=np.bool)
      self._stream_di.read_one_sample_multi_line(data)

      ret.extend(list(data[:, 0]))

    return ret

  def get_stream(self) -> Optional[List[np.ndarray]]:
    """Reads data from the NI DAQmx, and returns it in an array along with an
    array holding the timestamps.

    Only data from analog input channels can be read, this method cannot read
    stream data from digital input channels.
    """

    if not self._stream_started:
      return

    # Creating the container for the data
    data = np.empty((len(list(chain(*self._analog_in.values()))),
                     self._n_samples))
    # # Creating the array holding the timestamps
    t = time() + np.arange(0, self._n_samples) / self._sample_rate
    # Actually reading the data from the device
    for type_, stream in self._stream_ai.items():
      stream.read_many_sample(data, self._n_samples)

    return [t, data]

  def set_cmd(self, *cmd: float) -> None:
    """Sets the analog and digital output channels according to the given
    command values.

    The first command values correspond to the analog channels, the remaining
    ones correspond the the digital channels. It might be that not all channels
    are set if the number of commands doesn't match the number of channels.
    """

    # Setting the analog channels
    if self._analog_out:
      self._stream_ao.write_one_sample(np.array(cmd[:len(self._analog_out)],
                                       dtype=np.float64))

    # Setting the digital channels
    if self._digital_out:
      self._stream_do.write_one_sample_multi_line(
        np.array(cmd[len(self._analog_out):],
                 dtype=np.bool).reshape(len(self._digital_out), 1))

  def stop_stream(self) -> None:
    """Stops all the acquisition tasks."""

    if self._stream_started:
      for task in self._tasks_ai.values():
        task.stop()
        task.timing.cfg_samp_clk_timing(
          self._sample_rate,
          sample_mode=nidaqmx.constants.AcquisitionType.FINITE)

    self._stream_started = False

  def close(self) -> None:
    """Stops all the acquisition tasks, and closes the connections to the
    device."""

    if self._analog_in:
      for task in self._tasks_ai.values():
        task.stop()

    if self._analog_out:
      self._task_ao.stop()

    if self._stream_started:
      self.stop_stream()

    if self._analog_in:
      for task in self._tasks_ai.values():
        task.close()

    if self._analog_out:
      self._task_ao.close()



# coding: utf-8

from time import time
from numpy import loadtxt, interp
from typing import Dict, Union
import pathlib

from .path import Path


class Custom(Path):
  """Generates a custom path from a text file, until the file is exhausted.

  The file can be in any text format, including the most common `.csv` and
  `.txt` extensions.
  """

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               filename: Union[str, pathlib.Path],
               delimiter: str = ',') -> None:
    """Loads the file and sets the args.

    The stop condition is simply to reach the last timestamp given in the
    file.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      filename: Path to the file to read the path from. Can be either a
        :obj:`str` or a :mod:`pathlib` Path. The file must contain two columns:
        the first one containing timestamps (starting from 0), the other one
        containing the values.
      delimiter: The delimiter between columns in the file, usually a coma.
    """

    Path.__init__(self, _last_time, _last_cmd)

    array = loadtxt(pathlib.Path(filename), delimiter=delimiter)

    if array.shape[1] != 2:
      raise ValueError(f'The file {filename} should contain exactly two'
                       f'columns !')

    self._timestamps = array[:, 0]
    self._values = array[:, 1]

  def get_cmd(self, _: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met.

    The value is interpolated from the given file.
    """

    t = time()
    if t - self.t0 > self._timestamps[-1]:
      raise StopIteration
    return interp(t - self.t0, self._timestamps, self._values)



# coding: utf-8

"""
Example demonstrating the use of thermocouples with the Labjack T7 board.

required hardware:
  - Labjack T7 board
  - Thermocouple(s)
"""

import crappy

chan = [0, 2, 3, 4, 5]

if __name__ == "__main__":
  m = crappy.blocks.IOBlock("Labjack_t7",
                            channels=[dict(name='AIN%d' % i, thermocouple='K')
                                      for i in chan],
                            verbose=True, labels=['t(s)'] + ['T%d' % i
                                                             for i in chan])

  g = crappy.blocks.Grapher(*[('t(s)', 'T%d' % i) for i in chan])

  crappy.link(m, g, modifier=crappy.modifier.Moving_avg(10))

  crappy.start()



# coding: utf-8

from typing import Dict, Any, List
from time import time
from .inout import InOut
from .._global import OptionalModule

try:
  from pijuice import PiJuice
except (ModuleNotFoundError, ImportError):
  PiJuice = OptionalModule("pijuice")

try:
  from smbus2 import SMBus
except (ImportError, ModuleNotFoundError):
  SMBus = OptionalModule('smbus2')

pijuice_commands = {'status': 0x40,
                    'fault_event': 0x44,
                    'charge_level': 0x41,
                    'button_event': 0x45,
                    'battery_temperature': 0x47,
                    'battery_voltage': 0x49,
                    'battery_current': 0x4B,
                    'io_voltage': 0x4D,
                    'io_current': 0x4F,
                    'led_state': 0x66,
                    'led_blink': 0x68,
                    'io_pin_access': 0x75}

pijuice_battery_status = {0x00: 'Normal',
                          0x04: 'Charging_from_usb',
                          0x08: 'Charging_from_gpio',
                          0x0C: 'Absent'}

pijuice_power_status = {0x00: 'Not_present',
                        0x10: 'Bad',
                        0x20: 'Weak',
                        0x30: 'Present'}

pijuice_backends = ['Pi4', 'pijuice']


class Pijuice(InOut):
  """Block getting various information about a piJuice power platform,
  including the charge level and the power supply status.

  Warning:
    Only available on Raspberry Pi !
  """

  def __init__(self,
               i2c_port: int = 1,
               address: int = 0x14,
               backend: str = 'Pi4') -> None:
    """Checks arguments validity.

    Args:
      backend (:obj:`str`, optional): Should be one of :
        ::

          'Pi4', 'pijuice'

        The `'Pi4'` backend is based on the :mod:`smbus2` module, while the
        `'pijuice'` backend is based on the :mod:`pijuice`.
      i2c_port(:obj:`int`, optional): The I2C port over which the PiJuice
        should communicate.
      address(:obj:`int`, optional): The I2C address of the piJuice. The
        default address is 0x14.
    """

    super().__init__()
    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an int")
    self._i2c_port = i2c_port

    if not isinstance(address, int):
      raise TypeError("address should be an int")
    else:
      self._address = address

    if not isinstance(backend, str) or backend not in pijuice_backends:
      raise ValueError("backend should be in {}".format(pijuice_backends))
    self._backend = backend

  def open(self) -> None:
    """Opens the I2C port."""

    if self._backend == 'Pi4':
      self._bus = SMBus(self._i2c_port)
    elif self._backend == 'pijuice':
      self._pijuice = PiJuice(self._i2c_port, self._address)

  def get_data(self) -> Dict[str, Any]:
    """Reads all the available information on the battery status.

    Returns:
      :obj:`dict`: Returns a dict containing:

        - the timeframe in seconds as a :obj:`float` in label ``t(s)``
        - the battery status as a :obj:`str` in label ``battery_status``
        - the USB status as a :obj:`str` in label ``USB_status``
        - the GPIO status as a :obj:`str` in label ``GPIO_status``
        - the charge level as an :obj:`int` in label ``charge_level``
        - the battery temperature in °C as an :obj:`int` in label
          ``battery_temperature``
        - the battery voltage in mV as an :obj:`int` in label
          ``battery_voltage``
        - the battery current in mA as an :obj:`int` in label
          ``battery_current``
        - the GPIO voltage in mV as an :obj:`int` in label ``GPIO_voltage``
        - the GPIO current in mA as an :obj:`int` in label ``GPIO_current``
    """

    # Gets the battery status
    status = dict()

    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['status'], 2)
      status['battery'] = pijuice_battery_status[data[0] & 0x0C]
      status['USB'] = pijuice_power_status[data[0] & 0x30]
      status['GPIO'] = pijuice_power_status[data[0] & 0xC0]

    elif self._backend == 'pijuice':
      stat = self._pijuice.status.GetStatus()
      status['battery'] = stat['data']['battery']
      status['USB'] = stat['data']['powerInput']
      status['GPIO'] = stat['data']['powerInput5vIo']

    # Gets the battery charge level
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['charge_level'], 2)
      charge_level = data[0]
    else:
      data = self._pijuice.status.GetChargeLevel()
      charge_level = data['data']

    # Gets the battery temperature
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['battery_temperature'], 3)
      temp = data[0]
      if temp & 0x80:
        temp -= 0X100
    else:
      data = self._pijuice.status.GetBatteryTemperature()
      temp = data['data']

    # Gets the battery voltage (mV)
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['battery_voltage'], 3)
      voltage = (data[1] << 8) | data[0]
    else:
      data = self._pijuice.status.GetBatteryVoltage()
      voltage = data['data']

    # Gets the battery current (mA)
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['battery_current'], 3)
      current = (data[1] << 8) | data[0]
      if current & 0x8000:
        current -= 0x10000
    else:
      data = self._pijuice.status.GetBatteryCurrent()
      current = data['data']

    # Gets the GPIO voltage (mV)
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['io_voltage'], 3)
      io_voltage = (data[1] << 8) | data[0]
    else:
      data = self._pijuice.status.GetIoVoltage()
      io_voltage = data['data']

    # Gets the GPIO current (mA)
    if self._backend == 'Pi4':
      data = self._read_i2c(pijuice_commands['io_current'], 3)
      io_current = (data[1] << 8) | data[0]
      if io_current & 0x8000:
        io_current -= 0x10000
    else:
      data = self._pijuice.status.GetIoCurrent()
      io_current = data['data']

    return {'t(s)': time(),
            'battery_status': status['battery'],
            'USB_status': status['USB'],
            'GPIO_status': status['GPIO'],
            'charge_level': charge_level,
            'battery_temperature': temp,
            'battery_voltage': voltage,
            'battery_current': current,
            'GPIO_voltage': io_voltage,
            'GPIO_current': io_current}

  def close(self) -> None:
    """Closes the I2C bus."""

    if self._backend == 'Pi4':
      self._bus.close()

  @staticmethod
  def _checksum(data: List[int]) -> List[int]:
    """Compares the received checksum (last byte) to the one calculated over
    the other bytes.

    Also tries with the MSB of the first byte set to 1, as it is apparently
    sometimes 0 when it should be 1.

    Returns:
      The list of data bytes, with the MSB of the first byte corrected if
      needed.

    Raises:
      Raises an :exc:`IOError` if the checksums don't match.
    """

    # First checking on the raw data
    check = 0xFF
    for x in data[:-1]:
      check ^= x

    # If needed, modify the MSB of the first byte and retry
    if check != data[-1]:
      data[0] |= 0x80
      check = 0xFF
      for x in data[:-1]:
        check ^= x

      # If still not successful, raise error
      if check != data[-1]:
        raise IOError("An error occurred, the checksums don't match !")

    return data[:-1]

  def _read_i2c(self, command: int, length: int) -> list:
    """Thin wrapper to reduce verbosity."""

    ret = self._bus.read_i2c_block_data(i2c_addr=self._address,
                                        register=command,
                                        length=length)
    return self._checksum(ret)



# coding: utf-8

import tkinter as tk
from typing import Optional
from .cameraConfigBoxes import Camera_config_with_boxes
from .cameraConfigTools import Box


class VE_config(Camera_config_with_boxes):
  """Class similar to :ref:`Camera configuration` but also displaying the
  bounding boxes of the detected spots, and allowing to select the area where
  to detect the spots by drawing a box with the left mouse button.

  It is meant to be used for configuring the :ref:`VideoExtenso` block.
  """

  def __init__(self, camera, video_extenso) -> None:
    """Sets the args and initializes the parent class.

    Args:
      camera: The camera object in charge of acquiring the images.
      video_extenso: The video extenso tool in charge of tracking the spots.
    """

    self._video_extenso = video_extenso
    super().__init__(camera)

  def _bind_canvas_left_click(self) -> None:
    """Binds the left mouse button click for drawing the box in which the spots
    will be searched."""

    self._img_canvas.bind('<ButtonPress-1>', self._start_box)
    self._img_canvas.bind('<B1-Motion>', self._extend_box)
    self._img_canvas.bind('<ButtonRelease-1>', self._stop_box)

  def _create_buttons(self) -> None:
    """Compared with the parent class, creates an extra button for saving the
    original position of the spots."""

    self._update_button = tk.Button(self._sets_frame, text="Apply Settings",
                                    command=self._update_settings)
    self._update_button.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

    self._update_button = tk.Button(self._sets_frame, text="Save L0",
                                    command=self._save_l0)
    self._update_button.pack(expand=False, fill='none', ipadx=5, ipady=5,
                             padx=5, pady=5, anchor='n', side='top')

  def _stop_box(self, _: tk.Event) -> None:
    """When the user releases the mouse, searches for spots in the selected
    area and displays them if any were found."""

    # If it's just a regular click with no dragging, do nothing
    if self._img is None or self._select_box.no_points():
      self._select_box.reset()
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = self._select_box.sorted()

    # If the box is flat, resetting it
    if y_left == y_right or x_top == x_bottom:
      self._select_box.reset()
      return

    # Now actually trying to detect the spots
    try:
      spots = self._video_extenso.detect_spots(
          self._original_img[x_top: x_bottom, y_left: y_right], x_top, y_left)
      if spots is not None:
        self._spots = spots
    except IndexError:
      # Highly unlikely but always better to be careful
      self._spots.reset()
      return

    # This box is not needed anymore
    self._select_box.reset()

  def _save_l0(self) -> None:
    """Saves the original positions of the spots on the image."""

    if self._video_extenso.save_length():
      print(f"[VideoExtenso] Successfully saved L0 :\n"
            f"L0 x : {self._video_extenso.x_l0}\n"
            f"L0 y : {self._video_extenso.y_l0}")

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the patches and the
    select box on top of the displayed image."""

    self._draw_box(self._select_box)
    self._draw_spots()
    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the patches and the
    select box on top of the displayed image."""

    _, img = self._camera.get_image()
    if img is None:
      return

    self._cast_img(img)
    self._draw_box(self._select_box)
    self._draw_spots()
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, _: Box) -> None:
    """If a patch is outside the image, it means that the image size has been
    modified. Simply resetting the spots then."""

    self._spots.reset()



# coding: utf-8

"""
Example demonstrating the use of ``crappy.blocks.Multiplex``.

This block interpolates data from several parents in a common timebase.

No required hardware.
"""

import crappy


class Delay(crappy.modifier.Modifier):
  """Modifier to add a delay to one of the inputs, demonstrating how Multiplex
  will wait for data."""

  def __init__(self, n):
    super().__init__()
    self.n = n
    self.hist = []

  def evaluate(self, data):
    self.hist.append(data)
    if len(self.hist) >= self.n:
      return self.hist.pop(0)


if __name__ == "__main__":
  g1 = crappy.blocks.Generator([
    dict(type='sine', freq=1, amplitude=1, condition=None)
      ], freq=100, cmd_label='cmd1')

  g2 = crappy.blocks.Generator([
    dict(type='cyclic_ramp', speed2=-1, speed1=1,
         condition1='cmd2>1', condition2='cmd2<-1', cycles=0, init_value=0)
      ], freq=50, cmd_label='cmd2')

  mul = crappy.blocks.Multiplex(verbose=True)

  # crappy.link(g1, mul)
  crappy.link(g1, mul, modifier=Delay(50))
  crappy.link(g2, mul)

  graph = crappy.blocks.Grapher(('t(s)', 'cmd1'), ('t(s)', 'cmd2'))

  crappy.link(mul, graph)

  rec = crappy.blocks.Recorder("example_multi.csv",
                               labels=["t(s)", "cmd1", "cmd2"])

  crappy.link(mul, rec)

  crappy.start()



# coding: utf-8

from .block import Block
from .._global import OptionalModule
from struct import unpack
from time import time
try:
  from serial import Serial
  from serial.serialutil import SerialException
except (ModuleNotFoundError, ImportError):
  Serial = OptionalModule("pyserial")
  SerialException = OptionalModule("pyserial")


class UController(Block):
  """Block for interfacing over serial with an external device, written mostly
  for communication with microcontrollers.

  It can send labeled commands to the device, and/or receive labeled data from
  it. This block is meant to be used along with the MicroController.py
  MicroPython template located in the `tool` folder of Crappy, even though it
  is not mandatory. A given syntax needs to be followed for any data to be
  exchanged.
  """

  def __init__(self,
               labels: list = None,
               cmd_labels: list = None,
               init_output: dict = None,
               post_process: dict = None,
               t_device: bool = False,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 115200,
               verbose: bool = False,
               freq: float = 100) -> None:
    """Checks the validity of the arguments.

    Args:
      labels (:obj:`list`, optional): The list of the labels to get from the
        device. Only these labels should be given as argument to the
        :meth:`send_to_pc` method in the MicroPython script. If this argument
        is not :obj:`None`, then the``init_output`` argument should be given as
        well. No more than 9 labels should be given.
      cmd_labels (:obj:`list`, optional): The list of the command labels that
        will be sent to the device upon reception from an upstream block. The
        variables in the MicroPython script should have these exact names. No
        more than 9 cmd_labels should be given.
      init_output (:obj:`dict`, optional): If the ``labels`` argument is not
        :obj:`None`, the values to output to downstream blocks for each label
        as long as no value has been received from the device. An initial
        output value must be given for each label.
      post_process (:obj:`dict`, optional): Optionally allows applying a
        function to the data of a label before transmitting it to downstream
        blocks. It is possible to give functions for only part of the labels.
      t_device (:obj:`bool`, optional): It :obj:`True`, the timestamp returned
        under the label `'t(s)'` is the one of the device, not the one of
        Crappy. It may reduce the maximum achievable sample rate, as more bytes
        have to be transmitted, but it is also far more precise.
      port (:obj:`str`, optional): The serial port to open for communicating
        with the device. In Windows, they are usually called `COMx`, whereas in
        Linux and Mac they're called `/dev/ttyxxxx`.
      baudrate (:obj:`int`, optional): The baudrate for serial communication.
        It depends on the capabilities of the device.
      verbose (:obj:`bool`, optional): If :obj:`True`, prints debugging
        information.
      freq (:obj:`float`, optional): The looping frequency of the block.
    """

    super().__init__()

    if not isinstance(verbose, bool):
      raise TypeError("verbose should be either True or False !")
    self.verbose = verbose

    if not isinstance(freq, float) and not isinstance(freq, int) or freq <= 0:
      raise TypeError("freq should be a positive float !")
    self.freq = freq

    if not isinstance(t_device, bool):
      raise TypeError("t_device should be either True or False !")
    self._t_device = t_device

    if not isinstance(port, str):
      raise TypeError("port should be a string !")
    self._port = port

    if not isinstance(baudrate, int) or baudrate < 0:
      raise ValueError("baudrate should be a positive integer !")
    self._baudrate = baudrate

    if labels is not None and not isinstance(labels, list):
      raise TypeError('labels should be a list !')
    if labels is not None and len(labels) > 9:
      raise ValueError("Sorry, a maximum of 9 labels is allowed !")
    if cmd_labels is not None and not isinstance(cmd_labels, list):
      raise TypeError('cmd_labels should be a list !')
    if cmd_labels is not None and len(cmd_labels) > 9:
      raise ValueError("Sorry, a maximum of 9 cmd_labels is allowed !")
    self._labels = labels
    self._cmd_labels = cmd_labels

    if self._cmd_labels is not None:
      self._prev_cmd = {cmd_label: None for cmd_label in self._cmd_labels}

    if init_output is not None and not isinstance(init_output, dict):
      raise TypeError("init_output should be a dict !")
    if labels is not None and not all(label in (init_output if init_output
                                                is not None else {})
                                      for label in labels):
      raise ValueError("Every label should have an init_output value !")
    self._out = init_output

    if post_process is not None and (not isinstance(post_process, dict) or
                                     not all(callable(func) for func
                                             in post_process.values())):
      raise TypeError("post_process should be a dict of callables !")
    self._post_process = post_process if post_process is not None else {}

  def prepare(self) -> None:
    """Opens the serial port, and sends a `'go'` message to the device.

    Also shares with the device two tables associating each `cmd_label` and
    `label` with an integer. This allows reducing the traffic on the serial
    bus.

    Note:
      The commands are sent as text because some boards cannot read bytes from
      the `stdin` buffer in MicroPython.
    """

    # Checking if the link layout is relevant with respect to the arguments
    if self._labels is not None and not self.outputs:
      raise IOError("labels are specified but there's no output link !")
    if self._cmd_labels is not None and not self.inputs:
      raise IOError("cmd_labels are specified but there's no input link !")

    # Buffer for storing the received bytes
    if self._labels is not None:
      self._buffer = b''

    # Opening the serial port
    try:
      self._bus = Serial(self._port,
                         self._baudrate,
                         timeout=0,
                         write_timeout=0)
    except SerialException:
      raise IOError("Couldn't connect to the device on the port "
                    "{}".format(self._port))

    # Assigning indexes to the cmd_labels and labels, to identify them easily
    # and reduce the traffic on the bus
    self._cmd_table = {label: i for i, label in enumerate(self._cmd_labels,
                                                          start=1)} \
        if self._cmd_labels is not None else {}
    self._labels_table = {label: i for i, label in enumerate(self._labels,
                                                             start=1)} \
        if self._labels is not None else {}
    # The presence of the label 't(s)' indicates that the device should return
    # a timestamp along with the data
    if self._labels is not None and self._t_device:
      self._labels_table.update({'t(s)': 0})

    # Emptying the read buffer before starting
    while True:
      try:
        recv = self._bus.read()
      except SerialException:
        raise IOError("Reading from the device on port {} failed, it may"
                      "have been disconnected.".format(self._port))

      if not recv:
        break

    # Sending the 'go' command to start the device
    try:
      self._bus.write(b'go' +
                      str(len(self._cmd_table)).encode() +
                      str(len(self._labels_table)).encode() +
                      b'\r\n')
      if self.verbose:
        print('[UController] Sent {} on the port '
              '{}'.format(b'go' + str(len(self._cmd_table)).encode() +
                          str(len(self._labels_table)).encode() +
                          b'\r\n', self._port))
    except SerialException:
      raise IOError("Writing to the device on port {} failed, it may"
                    "have been disconnected.".format(self._port))

    # Sending the table of cmd_labels and their indexes
    for cmd, i in self._cmd_table.items():
      try:
        self._bus.write(str(i).encode() + cmd.encode() + b'\r\n')
        if self.verbose:
          print('[UController] Sent {} on the port {}'.format(
              str(i).encode() + cmd.encode() + b'\r\n', self._port))
      except SerialException:
        raise IOError("Writing to the device on port {} failed, it may"
                      "have been disconnected.".format(self._port))

    # Sending the table of labels and their indexes
    for label, i in self._labels_table.items():
      try:
        self._bus.write(str(i).encode() + label.encode() + b'\r\n')
        if self.verbose:
          print('[UController] Sent {} on the port {}'.format(
            str(i).encode() + label.encode() + b'\r\n', self._port))
      except SerialException:
        raise IOError("Writing to the device on port {} failed, it may"
                      "have been disconnected.".format(self._port))

  def loop(self) -> None:
    """First sends the commands from upstream blocks to the device, then reads
    the data from the device and sends it to the downstream blocks.

    Important:
      The precision of the commands sent to the device is limited to 3 digits
      after the decimal point, to limit the traffic on the bus. Adapt the range
      of the command values consequently.

    Note:
      Commands are sent as text, because some boards cannot read bytes from the
      `stdin` buffer in MicroPython. Data is however received on the PC from
      the device as bytes.
    """

    """Loop for sending the commands to the device.
    
    First, for each label in the cmd_labels list we search for a matching 
    command in the upstream links. Only the first one found is considered. If 
    the command value is different from the last value of this label, then the 
    command is sent to the device. Otherwise it is ignored. The command value
    is then stored as the last value of the label.
    """
    if self._cmd_labels is not None:
      received_cmd = [link.recv_chunk() if link.poll() else {} for link
                      in self.inputs]
      for label in self._cmd_labels:
        for dic in received_cmd:
          if label in dic:
            for value in dic[label]:
              if value != self._prev_cmd[label]:

                # Information for debugging
                if self.verbose:
                  print("[UController] Sent {} on the "
                        "port {}".format((str(self._cmd_table[label]) +
                                          str('%.3f' % value)).encode() +
                                         b'\r\n',
                                         self._port))

                # Sending the actual message to the device
                try:
                  self._bus.write((str(self._cmd_table[label]) +
                                   str('%.3f' % value)).encode() +
                                  b'\r\n')
                except SerialException:
                  raise IOError("Writing to the device on port {} failed, it "
                                "may have been "
                                "disconnected.".format(self._port))
                self._prev_cmd[label] = value
            break

    """Loop for receiving data from the device.
    
    A given number of bytes is read from the device, depending if `t_device` is
    True or False. they are then parsed to extract the index of the label, the
    data, and optionally the timestamp. If a label with a matching index is
    present in labels, its value is updated, as well as the timestamp. The
    values of ALL the labels are then sent to downstream blocks.
    """
    if self._labels is not None:
      # Reading the message from the device
      retries = 3
      while True:
        try:
          recv = self._bus.read()
        except SerialException:
          raise IOError("Reading from the device on port {} failed, it may"
                        "have been disconnected.".format(self._port))

        # This prevents the loop from staying stuck with an incomplete reading
        if not recv:
          retries -= 1
          if not retries:
           break
          continue

        self._buffer += recv
        # Exiting the loop when the desired number of bytes is reached
        if (self._t_device and
            len(self._buffer) == 9) or (not self._t_device and
                                        len(self._buffer) == 5):
          break

      # There was no message from the device, or it was incomplete
      if (self._t_device and
          len(self._buffer) != 9) or (not self._t_device and
                                      len(self._buffer) != 5):
        return

      # Information for debugging
      if self.verbose:
        print("[UController] Received {} on the port {}".format(self._buffer,
                                                                self._port))

      # Parsing the received bytes
      read = unpack('<ibf' if self._t_device else '<bf', self._buffer)
      self._buffer = b''

      # Reading the timestamp if relevant
      if self._t_device:
        self._out['t(s)'] = read[0] / 1000
        read = read[1:]

      # Updating the label value and sending to the downstream blocks
      for label in self._labels:
        if read[0] == self._labels_table[label]:
          value = read[1]
          self._out[label] = self._post_process[label](value) if \
              label in self._post_process else value
          if not self._t_device:
            self._out['t(s)'] = time() - self.t0
          self.send(self._out)
          break

    # Emptying the port buffer even if the messages are not processed
    else:
      while True:
        try:
          recv = self._bus.read()
        except SerialException:
          raise IOError("Reading from the device on port {} failed, it may"
                        "have been disconnected.".format(self._port))

        if not recv:
          break

        if self.verbose:
          print("[UController] Received {} on the port {}".format(recv,
                                                                  self._port))

  def finish(self) -> None:
    """Closes the serial port, and sends a `'stop!'` message to the device."""

    # Sending a 'stop!' message to the device
    try:
      self._bus.write(b'stop!' + b'\r\n')
      if self.verbose:
        print("[UController] Sent {} on the port {}".format(b'stop!' + b'\r\n',
                                                            self._port))
    except SerialException:
      pass

    self._bus.close()



# coding: utf-8

from typing import List
import tkinter as tk

from .block import Block


class Dashboard_window(tk.Tk):
  """The GUI for displaying the label values."""

  def __init__(self, labels: List[str]) -> None:
    """Initializes the GUI and sets the layout."""

    super().__init__()
    self.title('Dashboard')
    self.resizable(False, False)

    self._labels = labels

    # Attributes storing the tkinter objects
    self._tk_labels = {}
    self._tk_values = {}
    self.tk_var = {}

    # Setting the GUI
    self._set_variables()
    self._set_layout()

  def _set_variables(self) -> None:
    """Attributes one StringVar per label."""

    for label in self._labels:
      self.tk_var[label] = tk.StringVar(value='')

  def _set_layout(self) -> None:
    """Creates the Labels and places them on the GUI."""

    for row, label in enumerate(self._labels):
      # The name of the labels on the left
      self._tk_labels[label] = tk.Label(self, text=label, borderwidth=15,
                                        font=("Courier bold", 48))
      self._tk_labels[label].grid(row=row, column=0)
      # Their values on the right
      self._tk_values[label] = tk.Label(self, borderwidth=15,
                                        textvariable=self.tk_var[label],
                                        font=("Courier bold", 48))
      self._tk_values[label].grid(row=row, column=1)


class Dashboard(Block):
  """The Dashboard receives data from a :ref:`Link`, and prints it on a new
  popped window.

  It can only display data coming from one block.
  """

  def __init__(self,
               labels: List[str],
               nb_digits: int = 2,
               verbose: bool = False,
               freq: float = 30) -> None:
    """Sets the args and initializes parent class.

    Args:
      labels: Only the data from these labels will be printed on the window.
      nb_digits: Number of decimals to show.
      verbose: If :obj:`True`, prints the looping frequency of the block.
      freq: If set, the block will try to loop at this frequency.
    """

    super().__init__()
    self.verbose = verbose
    self.freq = freq

    self._labels = labels
    self._nb_digits = nb_digits

  def prepare(self) -> None:
    """Checks that there's only one incoming link, and starts the GUI."""

    if len(self.inputs) == 0:
      raise IOError("No link pointing towards the Dashboard block !")
    elif len(self.inputs) > 1:
      raise IOError("Too many links pointing towards the Dashboard block !")
    self._link, = self.inputs

    self._dashboard = Dashboard_window(self._labels)
    self._dashboard.update()

  def loop(self) -> None:
    """Receives the data from the incoming link and displays it."""

    data = self._link.recv_last()

    if data is not None:
      for label, value in data.items():
        # Only print the required labels
        if label in self._labels:
          # Possibility to display str values carried by the links
          if isinstance(value, str):
            self._dashboard.tk_var[label].set(value)
          elif isinstance(value, int) or isinstance(value, float):
            self._dashboard.tk_var[label].set(f'{value:.{self._nb_digits}f}')

    # In case the GUI has been destroyed, don't raise an error
    try:
      self._dashboard.update()
    except tk.TclError:
      pass

  def finish(self) -> None:
    """"""

    # In case the GUI has been destroyed, don't raise an error
    try:
      self._dashboard.destroy()
    except tk.TclError:
      pass



# coding: utf-8

import numpy as np
from typing import Union, Optional
from pathlib import Path

from .._global import OptionalModule

try:
  import tables
except ModuleNotFoundError:
  tables = OptionalModule("tables", "Hdf_recorder needs the tables module to "
                          "write hdf files.")

from .block import Block


class Hdf_recorder(Block):
  """This block saves data efficiently into a hdf5 file.

  This block is meant to save data coming as arrays at a high rate
  (`>1kHz`). It relies on the module :mod:`tables`.

  Important:
    Do not forget to specify the type of data to be saved (see ``atom``
    parameter) to avoid casting the data into another type, as this could
    result in data loss or inefficient saving.
  """

  def __init__(self,
               filename: Union[str, Path],
               node: str = 'table',
               expected_rows: int = 10**8,
               atom=None,
               label: str = 'stream',
               metadata: Optional[dict] = None,
               freq: Optional[float] = None,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      filename: Path to the output file, either relative or absolute. If the
        parent folders of the file do not exist, they will be created. If the
        file already exists, the actual file where data will be written will be
        renamed with a trailing index to avoid overriding it.
      expected_rows: The number of expected rows in the file. It is used to
        optimize the dumping.
      atom: This represents the type of data to be stored in the table. It can
        be given as a :class:`tables.Atom` instance, as a :class:`numpy.array`
        or as a :obj:`str`.
      label: The label carrying the data to be saved
      metadata: A :obj:`dict` containing additional information to save in the
        `hdf5` file.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.
    """

    Block.__init__(self)
    if freq is not None:
      self.freq = freq
    self.verbose = verbose

    self._path = Path(filename)
    self._label = label
    self._metadata = {} if metadata is None else metadata
    self._expected_rows = expected_rows

    self._node = node
    atom = tables.Int16Atom() if atom is None else atom
    if not isinstance(atom, tables.Atom):
      self._atom = tables.Atom.from_dtype(np.dtype(atom))
    else:
      self._atom = atom

  def prepare(self) -> None:
    """Checking that the block has the right number of inputs, creates the
    folder containing the file if it doesn't already exist, changes the name of
    the file if it already exists, and initializes the HDF file."""

    # Making sure there's the right number of incoming links
    if not self.inputs:
      raise ValueError('The HDF Recorder block does not have inputs !')
    elif len(self.inputs) > 1:
      raise ValueError('Cannot link more than one block to an HDF Recorder '
                       'block !')
    self._link = self.inputs[0]

    parent_folder = self._path.parent

    # Creating the folder for storing the data if it does not already exist
    if not Path.is_dir(parent_folder):
      Path.mkdir(parent_folder, exist_ok=True, parents=True)

    # Changing the name of the file if it already exists
    if Path.exists(self._path):
      print(f'[HDF Recorder] Warning ! The file {self._path} already exists !')
      stem, suffix = self._path.stem, self._path.suffix
      i = 1
      # Adding an integer at the end of the name to identify the file
      while Path.exists(parent_folder / f'{stem}_{i:05d}{suffix}'):
        i += 1
      self._path = parent_folder / f'{stem}_{i:05d}{suffix}'
      print(f'[HDF Recorder] Using {self._path} instead !')

    # Initializing the file to save data to
    self._hfile = tables.open_file(str(self._path), "w")
    for name, value in self._metadata.items():
      self._hfile.create_array(self._hfile.root, name, value)

  def begin(self) -> None:
    """Receives the first chunk of data, makes sure that it contains the label
    to save, and initializes the HDF array with it."""

    data = self._link.recv_chunk(blocking=True)

    if self._label not in data:
      raise KeyError(f'The data received by the HDF Recorder block does not '
                     f'contain the label {self._label} !')

    _, width, *_ = data[self._label][0].shape
    self._array = self._hfile.create_earray(self._hfile.root,
                                            self._node,
                                            self._atom,
                                            (0, width),
                                            expectedrows=self._expected_rows)
    for elt in data[self._label]:
      self._array.append(elt)

  def loop(self) -> None:
    """Simply receives data from the upstream block and saves it."""

    data = self._link.recv_chunk(blocking=False)

    if data is not None:
      for elt in data[self._label]:
        self._array.append(elt)

  def finish(self) -> None:
    """Simply closes the HDF file."""

    self._hfile.close()


class Hdf_saver(Hdf_recorder):
  def __init__(self, *args, **kwargs) -> None:
    print('#### WARNING ####\n'
          'The block "Hdf_saver" has been renamed to "Hdf_recorder".\n'
          'Please replace the name in your program, '
          'it will be removed in future versions\n'
          '#################')
    Hdf_recorder.__init__(self, *args, **kwargs)



# coding: utf-8

from time import time, sleep
from typing import Dict, List, Any, Optional, Iterator
from itertools import cycle
from copy import deepcopy

from .block import Block
from . import generator_path
from .._global import CrappyStop


class _GeneratorStop(Exception):
  """A custom exception for handling the case when the Generator should not
  raise a :exc:`CrappyStop` exception when it terminates."""


class Generator(Block):
  """This block generates a signal following a user-defined path.

  One Generator block can only generate one signal. Use multiple blocks if
  several signals are needed. Note that the default behavior of a Generator
  is to stop the entire script when it ends.

  It should be used as an input for other blocks, for example for driving a
  :ref:`Machine` or an :ref:`IOBlock`. It can also accept inputs, to make the
  path dependent on data coming from other blocks.
  """

  def __init__(self,
               path: List[Dict[str, Any]],
               freq: float = 200,
               cmd_label: str = 'cmd',
               cycle_label: str = 'cycle',
               repeat: bool = False,
               spam: bool = False,
               verbose: bool = False,
               end_delay: Optional[float] = 2,
               safe_start: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      path: It must be a :obj:`list` of :obj:`dict`, each dict providing the
        parameters to generate the path. Refer to the Note below for more
        information.
      freq: The looping frequency this block will try to achieve. Note that the
        higher this value, the more accurate the path will be. It will also
        consume more resources.
      cmd_label: The label of the signal sent to the downstream blocks.
      cycle_label: In addition to the `cmd_label`, this label holds the index
        of the current dict in the ``path`` list. Useful to trig a block  upon
        change in the current dict.
      repeat: If :obj:`True`, the ``path`` will loop forever instead of
        stopping when the list of dicts is exhausted.
      spam: If :obj:`True`, the signal value will be sent on each loop. Else,
        it will only be sent if it is different from the previous or if the
        ``path`` switched to a new dict.
      verbose: if :obj:`True`, displays the loop frequency of the block and a
        message when switching to the next dict of ``path``.
      end_delay: When all the dicts in ``path`` are exhausted, waits this many
        seconds before stopping the entire program. Can be set to :obj:`None`,
        in which case the Generator won't stop the program when finishing.
      safe_start: Ensures the first dict in ``path`` waits for at least one
        data point from upstream blocks before sending the first value of the
        signal. Otherwise, the first value might be sent without checking the
        associated condition if its depends on labels from other blocks.

    Note:
      The different types of signals that can be generated by the Generator
      can be found at `generator path`. The ``path`` list contains one dict per
      signal shape to generate. They are generated in the order in which they
      appear in the list.

      Each dict contains information on the signal shape to generate, like its
      type, any applicable parameter(s), and the stop condition(s). Refer to
      the documentation of each signal shape to which information to give.
    """

    Block.__init__(self)

    # Instantiating a few attributes based on the arguments
    self.niceness = -5
    self.freq = freq
    self.verbose = verbose
    self.labels = ['t(s)', cmd_label, cycle_label]
    self._end_delay = end_delay
    self._spam = spam
    self._safe_start = safe_start

    # Checking the validity of the path
    try:
      self._check_path_validity(iter(deepcopy(path)))
    except (Exception,):
      print("Error while parsing the Generator path !")
      raise

    # The path is an iterable object
    self._path = cycle(path) if repeat else iter(path)

    # More attributes
    self._ended_no_raise = False
    self._last_cmd = None
    self._last_id = None
    self._last_t = None
    self._current_path = None
    self._path_id = None

  def begin(self) -> None:
    """Initializes the first path and runs a :meth:`loop`, that may be
    blocking."""

    self._update_path()
    self.loop(blocking=self._safe_start)

  def loop(self, blocking: bool = False) -> None:
    """First reads data from upstream blocks, then gets the next command to
    send, and finally sends it to downstream blocks.

    It also manages the transitions between the paths.

    Args:
      blocking: It :obj:`True`, waits blocks until there's data available from
        the upstream blocks before getting the next command to send.
    """

    # Case when the Generator shouldn't raise CrappyStop after it ended
    if self._ended_no_raise:
      return

    # Getting the data from upstream blocks
    data = self.get_all_last(blocking=blocking)
    try:
      # Getting the next command to send
      self._last_t = time()
      cmd = self._current_path.get_cmd(data)
    except StopIteration:
      try:
        # Switching to the next path if we reached the end of one
        self._update_path()
        self.loop()
        return
      except _GeneratorStop:
        # Case when the Generator shouldn't raise CrappyStop after it ended
        self._ended_no_raise = True
        return

    # The command is sent if it's different from the previous, or if we
    # switched to the next path, or if spam is True
    if cmd != self._last_cmd or self._last_id != self._path_id or self._spam:
      self._last_cmd = cmd
      self._last_id = self._path_id
      # Actually sending the command
      self.send([self._last_t - self.t0, cmd, self._path_id])

  def _update_path(self) -> None:
    """Gets the next path from the list of paths and instantiates it.

    Also manages the case when the last path of the list was reached.
    """

    try:
      # Getting the next path from the list of paths
      next_path_dict = deepcopy(self._path.__next__())
      # Updating the path index
      if self._path_id is not None:
        self._path_id += 1
      else:
        self._path_id = 0

    except StopIteration:
      # Raised when the list of paths is exhausted
      print("Signal generator terminated !")
      # First option, stopping the program after a delay
      if self._end_delay is not None:
        sleep(self._end_delay)
        raise CrappyStop("Signal Generator terminated")
      # Second option, not stopping the program and looping forever
      else:
        raise _GeneratorStop

    # Warning the user that the path ended if spam is True
    if self.verbose:
      print(f"[Signal Generator] Next step({self._path_id}): {next_path_dict}")

    # Instantiating the next generator path object
    path_name = next_path_dict.pop('type').capitalize()
    path_type = getattr(generator_path, path_name)
    self._current_path = path_type(
      _last_time=self._last_t if self._last_t is not None else self.t0,
      _last_cmd=self._last_cmd,
      **next_path_dict)

  @staticmethod
  def _check_path_validity(path: Iterator[Dict[str, Any]]) -> None:
    """Simply instantiates all the paths in a row to check no error is
    raised."""

    for i, next_dict in enumerate(path):
      next_dict = deepcopy(next_dict)
      path_name = next_dict.pop('type').capitalize()
      path_type = getattr(generator_path, path_name)
      path_type(_last_time=0, _last_cmd=None if i == 0 else 0, **next_dict)



# coding: utf-8

"""
Demonstrates how to use the stream mode on Labjack T7 boards.

Hardware required:
  - Labjack T7
"""

import crappy
import tables


if __name__ == "__main__":
  s = crappy.blocks.IOBlock("T7_streamer",
                            channels=[{'name': 'AIN0', 'gain': 2,
                                       'offset': -13},
                                      {'name': 'AIN1', 'gain': 2,
                                       "make_zero": True}],
                            streamer=True)

  g = crappy.blocks.Grapher(('t(s)', 'AIN0'), ('t(s)', 'AIN1'))
  crappy.link(s, g, modifier=crappy.modifier.Demux(labels=('AIN0', 'AIN1')))

  rec = crappy.blocks.Hdf_recorder("out.h5", atom=tables.Float64Atom())
  crappy.link(s, rec)
  crappy.start()



# coding: utf-8

import crappy


help_string = """
Syntax:
{axe}{speed}
(You can put a space between axe and speed)
Example:
  x -50
Will pull the axe in at 50% of the speed
  a20
Will push out motor A only (the one with numdevice=1)

Use q to quit
Use clear to clear the errors on all the motors
Type 'h' or 'help' to see this message again
"""


actuator_list = list()
actuator_list.append(crappy.actuator.Oriental(port='/dev/ttyUSB0'))
actuator_list.append(crappy.actuator.Oriental(port='/dev/ttyUSB1'))
actuator_list.append(crappy.actuator.Oriental(port='/dev/ttyUSB2'))
actuator_list.append(crappy.actuator.Oriental(port='/dev/ttyUSB3'))
for act in actuator_list:
  act.open()

motors = ['A', 'B', 'C', 'D']
actuator_list = sorted(actuator_list, key=lambda x: x.num_device)
actuator_dict = {}
i = 0
for c in 'abcd':
  actuator_dict[c] = [actuator_list[i]]
  i += 1
actuator_dict['x'] = actuator_dict['b']+actuator_dict['d']
actuator_dict['y'] = actuator_dict['a']+actuator_dict['c']

actuator_dict['z'] = actuator_dict['x']+actuator_dict['y']

print(help_string)
for i in range(4):
  print("BiaxeTech{} (on {}) is motor {}".format(
            i, actuator_list[i].port, motors[actuator_list[i].num_device - 1]))

user_input = ''
while True:
  print("Axe and speed ?>", end="")
  try:
    user_input = input().lower()
  except (EOFError, KeyboardInterrupt):
    break
  if user_input == 'q':
    break
  elif user_input in ['h', 'help', '?']:
    print(help_string)
    continue
  elif user_input == 'clear':
    for axe in actuator_list:
      axe.clear_errors()
    continue
  try:
    axe = user_input[:1]
    speed = int(user_input[1:].strip()) * .07  # From % to mm/min
  except ValueError:
    print("Unknown command, stopping")
    for axe in actuator_list:
      axe.set_speed(0)
    continue

  if user_input[0] in 'xyzabcd':
    for axe in actuator_dict[user_input[0]]:
      axe.set_speed(speed)

for axe in actuator_list:
  axe.set_speed(0)
  axe.close()



# coding: utf-8

from time import time
from typing import Tuple
from numpy import ndarray
from platform import system
from subprocess import run
from re import findall, split, search
from .camera import Camera
from .._global import OptionalModule

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")

# Todo: Manage frame rate


class Camera_opencv(Camera):
  """A class for reading images from any camera able to interface with OpenCv.

  The number of the video device to read images from can be specified. It is
  then also possible to tune the encoding format and the size.

  This camera class is less performant than the :ref:`Camera GStreamer` one
  that relies on GStreamer, but the installation of OpenCv is way easier than
  the one of GStreamer.

  Note:
    For a better performance of this class in Linux, it is recommended to have
    `v4l-utils` installed.
  """

  def __init__(self) -> None:
    """Sets variables and adds the channels setting."""

    super().__init__()
    self._cap = None

    self.add_choice_setting(name="channels",
                            choices=('1', '3'),
                            default='1')

  def open(self, device_num: int = 0, **kwargs) -> None:
    """Opens the video stream and sets any user-specified settings.

    Args:
      device_num (:obj:`int`, optional): The number of the device to open.
      **kwargs: Any additional setting to set before opening the graphical
        interface.
    """

    if not isinstance(device_num, int) or device_num < 0:
      raise ValueError("device_num should be an integer !")

    # Opening the videocapture device
    self._cap = cv2.VideoCapture(device_num)
    self._device_num = device_num
    fourcc = self._get_fourcc()

    if system() == 'Linux':
      self._formats = []

      # Trying to run v4l2-ctl to get the available formats
      command = ['v4l2-ctl', '-d', str(device_num), '--list-formats-ext']
      try:
        check = run(command, capture_output=True, text=True)
      except FileNotFoundError:
        print("\n#######\n"
              "Warning ! The performance of the Camera_opencv "
              "class could be improved if v4l-utils was installed !"
              "\n#######\n")
        check = None
      check = check.stdout if check is not None else ''

      # Splitting the returned string to isolate each encoding
      if findall(r'\[\d+]', check):
        check = split(r'\[\d+]', check)[1:]
      elif findall(r'Pixel\sFormat', check):
        check = split(r'Pixel\sFormat', check)[1:]
      else:
        check = []

      if check:
        for img_format in check:
          # For each encoding, finding its name
          name, *_ = search(r"'(\w+)'", img_format).groups()
          sizes = findall(r'\d+x\d+', img_format)

          # For each name, finding the available sizes
          for size in sizes:
            self._formats.append(f'{name} {size}')

      else:
        # If v4l-utils is not installed, proposing two encodings without
        # further detail
        self._formats = [fourcc, 'MJPG']

        # Still letting the user choose the size
        self.add_scale_setting(name='width', lowest=1, highest=1920,
                               getter=self._get_width, setter=self._set_width)
        self.add_scale_setting(name='height', lowest=1, highest=1080,
                               getter=self._get_height,
                               setter=self._set_height)

    else:
      # On Windows the fourcc management is even messier than on Linux
      self._formats = []

      # Still letting the user choose the size
      self.add_scale_setting(name='width', lowest=1, highest=1920,
                             getter=self._get_width, setter=self._set_width)
      self.add_scale_setting(name='height', lowest=1, highest=1080,
                             getter=self._get_height, setter=self._set_height)

    if self._formats:
      # The format integrates the size selection
      if ' ' in self._formats[0]:
        self.add_choice_setting(name='format',
                                choices=tuple(self._formats),
                                getter=self._get_format_size,
                                setter=self._set_format)
      # The size is independent of the format
      else:
        self.add_choice_setting(name='format',
                                choices=tuple(self._formats),
                                getter=self._get_fourcc,
                                setter=self._set_format)

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Grabs a frame from the videocapture object and returns it along with a
    timestamp."""

    # Grabbing the frame and the timestamp
    t = time()
    ret, frame = self._cap.read()

    # Checking the integrity of the frame
    if not ret:
      raise IOError("Error reading the camera")

    # Returning the image in the right format, and its timestamp
    if self.channels == '1':
      return t, cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    else:
      return t, frame

  def close(self) -> None:
    """Releases the videocapture object."""

    if self._cap is not None:
      self._cap.release()

  def _get_fourcc(self) -> str:
    """Returns the current fourcc string of the video encoding."""

    fcc = int(self._cap.get(cv2.CAP_PROP_FOURCC))
    return f"{fcc & 0xFF:c}{(fcc >> 8) & 0xFF:c}" \
           f"{(fcc >> 16) & 0xFF:c}{(fcc >> 24) & 0xFF:c}"

  def _get_width(self) -> int:
    """Returns the current image width."""

    return self._cap.get(cv2.CAP_PROP_FRAME_WIDTH)

  def _get_height(self) -> int:
    """Returns the current image height."""

    return self._cap.get(cv2.CAP_PROP_FRAME_HEIGHT)

  def _set_width(self, width: int) -> None:
    """Tries to set the image width."""

    self._cap.set(cv2.CAP_PROP_FRAME_WIDTH, width)

  def _set_height(self, height: int) -> None:
    """Tries to set the image height."""

    self._cap.set(cv2.CAP_PROP_FRAME_HEIGHT, height)

  def _set_format(self, img_format: str) -> None:
    """Sets the format of the image according to the user's choice."""

    # The format might be made of a name and a dimension, or just a name
    try:
      format_name, img_size = img_format.split(' ')
    except ValueError:
      format_name, img_size = img_format, None

    # Setting the format
    self._cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*format_name))

    if img_size is not None:
      # Getting the width and height from the second half of the string
      width, height = map(int, img_size.split('x'))

      # Setting the size
      self._set_width(width),
      self._set_height(height)

  def _get_format_size(self) -> str:
    """Parses the v4l2-ctl -V command to get the current image format as an
    index."""

    # Sending the v4l2-ctl command
    command = ['v4l2-ctl', '-d', str(self._device_num), '-V']
    check = run(command, capture_output=True, text=True).stdout

    # Parsing the answer
    format_ = width = height = ''
    if search(r"'(\w+)'", check) is not None:
      format_, *_ = search(r"'(\w+)'", check).groups()
    if search(r"(\d+)/(\d+)", check):
      width, height = search(r"(\d+)/(\d+)", check).groups()

    return f'{format_} {width}x{height}'



# coding: utf-8

from time import sleep
from .actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Biaxe(Actuator):
  """This class creates an axis and opens the corresponding serial port."""

  def __init__(self,
               port: str = '/dev/ttyUSB0',
               baudrate: int = 38400,
               timeout: float = 1) -> None:
    """Sets the instance attributes.

    Args:
      port (:obj:`str`, optional): Path to the corresponding serial port, e.g
        `'/dev/ttyS4'`.
      baudrate (:obj:`int`, optional): Set the corresponding baud rate.
      timeout (:obj:`float`, optional): Serial timeout.
    """

    Actuator.__init__(self)
    self.port = port
    self.baudrate = baudrate
    self.timeout = timeout

  def open(self) -> None:
    self.ser = serial.Serial(self.port, self.baudrate,
                             serial.EIGHTBITS, serial.PARITY_EVEN,
                             serial.STOPBITS_ONE, self.timeout)
    self.clear_errors()
    self.speed = None

  def stop(self) -> None:
    self.set_speed(0)

  def close(self) -> None:
    """Close the designated port."""

    self.stop()
    sleep(.01)
    self.ser.close()

  def clear_errors(self) -> None:
    """Reset errors."""

    self.ser.write(b"CLRFAULT\r\n")
    self.ser.write(b"OPMODE 0\r\n EN\r\n")

  def set_speed(self, speed: float) -> None:
    """Re-define the speed of the motor.
    ::

      1 = 0.002 mm/s

    """

    s = int(speed / .002)  # Convert to mm/s
    if s != self.speed:  # If it changed since last time (to avoid spamming)
      self.ser.write(b"J " + str(s).encode('ASCII') + b"\r\n")
      self.speed = s



# coding: utf-8

from time import time
from typing import Dict, List, Any, Optional

from .block import Block
from ..actuator import actuator_list


class Machine(Block):
  """This block is meant to drive one or several :ref:`Actuators`.

  The possibility to drive several Actuators from a unique block is given so
  that they can be driven in a synchronized way. If synchronization is not
  needed, it is preferable to drive the Actuators from separate Machine blocks.
  """

  def __init__(self,
               actuators: List[Dict[str, Any]],
               common: Optional[Dict[str, Any]] = None,
               time_label: str = 't(s)',
               spam: bool = False,
               freq: float = 200,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      actuators: The :obj:`list` of all the :ref:`Actuators` this block needs
        to drive. It contains one :obj:`dict` for every Actuator, with
        mandatory and optional keys. The keys providing information on how to
        drive the Actuator are listed below. Any other key will be passed to
        the Actuator object as argument when instantiating it.
      common: The keys of this :obj:`dict` will be common to all the Actuators.
        If it conflicts with an existing key for an Actuator, the common one
        will prevail.
      time_label: If reading speed or position from one or more Actuators, the
        time information will be carried by this label.
      spam: If :obj:`True`, a command is sent to the Actuators on each loop of
        the block, else it is sent every time a command is received.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.

    Note:
      - ``actuators`` keys:

        - ``type``: The name of the Actuator class to instantiate.
        - ``cmd``: The label carrying the command for driving the Actuator.
        - ``mode``: Can be either `'speed'` or `'position'`. Will either call
          :meth:`set_speed` or :meth:`set_position` to drive the actuator, and
          :meth:`get_speed` or :meth:`get_position` for acquiring the current
          speed or position.
        - ``speed``: If mode is `'position'`, the speed at which the Actuator
          should move. This key is not mandatory, even in the `'position'`
          mode.
        - ``pos_label``: If given and the mode is `'position'`, the block will
          return the value of :meth:`get_position` under this label. This key
          is not mandatory.
        - ``speed_label``: If given and the mode is `'speed'`, the block will
          return the value of :meth:`get_speed` under this label. This key is
          not mandatory.
    """

    super().__init__()
    self.freq = freq
    self.verbose = verbose

    if common is None:
      common = dict()

    self._time_label = time_label
    self._spam = spam

    self._settings = list()

    # For each actuator, parsing the given settings
    for actuator in actuators:
      actuator.update(common)
      settings = dict()

      # Getting all the possible arguments from the actuator dict
      for arg in ('type', 'cmd', 'mode', 'speed', 'pos_label', 'speed_label'):
        try:
          settings[arg] = actuator.pop(arg)
        except KeyError:
          # If an arg is not given, setting it to None
          settings[arg] = None

      # Putting all the remaining settings together under the kwargs key
      settings['kwargs'] = actuator

      # Making sure that the mandatory arguments are given
      for arg in ('type', 'cmd', 'mode'):
        if settings[arg] is None:
          raise ValueError(f"An actuator given as argument of the Machine "
                           f"block doesn't define the {arg} setting !")

      # Making sure that the given mode is valid
      if settings['mode'].lower() not in ('position', 'speed'):
        raise ValueError(f"The 'mode' setting for the actuators should be "
                         f"either 'position' or 'speed' !")

      # Storing the settings dict
      self._settings.append(settings)

    # Instantiating the actuators
    self._actuators = [actuator_list[settings['type'].capitalize()]
                       (**settings['kwargs']) for settings in self._settings]

  def prepare(self) -> None:
    """Checks the validity of the linking and initializes all the Actuator
    objects to drive."""

    # Checking the consistency of the linking
    if not self.inputs and not self.outputs:
      raise IOError("The Machine block is neither an input nor an output !")

    # Opening each actuator
    for actuator in self._actuators:
      actuator.open()

  def loop(self) -> None:
    """Receives the commands from upstream blocks, sets them on the actuators
    to drive, and sends the read positions and speed to the downstream
    blocks."""

    # Receiving the latest command
    if self._spam:
      recv = self.get_last(blocking=False)
    else:
      recv = self.recv_all_last()

    # Iterating over the actuators for setting the commands
    if recv:
      for actuator, settings in zip(self._actuators, self._settings):
        mode, cmd = settings['mode'], settings['cmd']

        # If a command was received, setting it
        if cmd in recv:
          if mode == 'speed':
            actuator.set_speed(recv[cmd])
          elif mode == 'position':
            actuator.set_position(recv[cmd], settings['speed'])

    to_send = {}

    # Iterating over the actuators to get the speeds and the positions
    for actuator, settings in zip(self._actuators, self._settings):
      pos_label, speed_label = settings['pos_label'], settings['speed_label']
      mode = settings['mode']
      if mode == 'position' and pos_label is not None:
        position = actuator.get_position()
        if position is not None:
          to_send[pos_label] = position
      elif mode == 'speed' and speed_label is not None:
        speed = actuator.get_speed()
        if speed is not None:
          to_send[speed_label] = speed

    # Sending the speed and position values if any
    if to_send:
      to_send[self._time_label] = time() - self.t0
      self.send(to_send)

  def finish(self) -> None:
    """Stops and closes all the actuators to drive."""

    for actuator in self._actuators:
      actuator.stop()
    for actuator in self._actuators:
      actuator.close()



# coding: utf-8

import numpy as np
import tkinter as tk
from .cameraConfig import Camera_config
from .cameraConfigTools import Box, Spot_boxes


class Camera_config_with_boxes(Camera_config):
  """This class is a basis for the configuration GUIs featuring boxes to
  display or to draw.

  It implements useful methods for drawing the boxes. If instantiated, this
  class behaves the exact same way as its parent class. It is not used as is by
  any block in Crappy.
  """

  def __init__(self, camera) -> None:
    """Initializes the parent class and sets the spots container."""

    self._spots = Spot_boxes()
    self._select_box = Box()
    super().__init__(camera)

  def _draw_box(self, box: Box) -> None:
    """Draws one line of the box after the other, making sure they fit in the
    image."""

    if self._img is None or box.no_points():
      return

    # The sides need to be sorted before slicing numpy array
    y_left, y_right, x_top, x_bottom = box.sorted()

    # Drawing one line after the other
    for slice_ in ((slice(x_top, x_bottom), y_left),
                   (slice(x_top, x_bottom), y_right),
                   (x_top, slice(y_left, y_right)),
                   (x_bottom, slice(y_left, y_right))):
      try:
        # The color of the line is adjusted according to the background
        # The original image must be used as no lines are already drawn on it
        if np.size(self._original_img[slice_]) > 0:
          self._img[slice_] = 255 * np.rint(np.mean(self._img[slice_] < 128))
      except IndexError:
        self._handle_box_outside_img(box)
        return

  def _handle_box_outside_img(self, _: Box) -> None:
    """This method is meant to simplify the customization of the action to
    perform when a patch is outside the image in subclasses."""

    pass

  def _draw_spots(self) -> None:
    """Simply draws every spot on top of the image."""

    if self._img is None:
      return

    for spot in self._spots:
      if spot is not None:
        self._draw_box(spot)

  def _start_box(self, event: tk.Event) -> None:
    """Simply saves the position of the user click."""

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self._select_box.x_start, \
        self._select_box.y_start = self._coord_to_pix(event.x, event.y)

  def _extend_box(self, event: tk.Event) -> None:
    """Draws a box as the user drags the mouse while maintaining the left
    button clicked."""

    # If the mouse is on the canvas but not on the image, do nothing
    if not self._check_event_pos(event):
      return

    self._select_box.x_end, \
        self._select_box.y_end = self._coord_to_pix(event.x, event.y)



# coding: utf-8

from time import time
from typing import Dict, Any, Optional

from .block import Block
from ..actuator import actuator_list


class AutoDrive(Block):
  """This block is meant to drive an actuator on which a camera performing
  videoextensometry is mounted so that the spots stay centered on the image.

  It takes the output of a :ref:`VideoExtenso` block and uses the coordinates
  of the spots to drive the actuator. The actuator can only be driven in speed,
  not in position.

  It also outputs the difference between the center of the image and the middle
  of the spots, along with a timestamp, over the ``'t(s)'`` and ``'diff(pix)'``
  labels. It can then be used by downstream blocks.
  """

  def __init__(self,
               actuator: Optional[Dict[str, Any]] = None,
               gain: float = 2000,
               direction: str = 'Y-',
               pixel_range: int = 2048,
               max_speed: float = 200000,
               freq: float = 200,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      actuator: A :obj:`dict` for initializing the actuator to drive. It
        should contain the name of the actuator under the key ``'name'``, and
        all the arguments to pass to the actuator as key/value pairs. The
        default actuator if this argument is not set is the :ref:`CM Drive`
        with its default arguments.
      gain: The gain for driving the actuator in speed. The speed command is
        simply the difference in pixels between the center of the image and the
        center of the spots, multiplied by this gain.
      direction: Indicates which axis to consider for driving the actuator, and
        whether the action should be inverted. The first character is the axis
        (`X` or `Y`) and second character is the inversion (`+` or `-`). The
        inversion depends on whether a positive speed will make bring the spots
        closer or farther, you have to try !
      pixel_range: The size of the image (in pixels) along the chosen axis.
      max_speed: The absolute maximum speed value that can be sent to the
        actuator.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.
    """

    super().__init__()
    self.labels = ['t(s)', 'diff(pix)']
    self.freq = freq
    self.verbose = verbose

    self._actuator = {'name': 'CM_drive'} if actuator is None else actuator
    self._gain = -gain if '-' in direction else gain
    self._direction = direction
    self._pixel_range = pixel_range
    self._max_speed = max_speed

  def prepare(self) -> None:
    """Checks the consistency of the linking and initializes the actuator to
    drive."""

    # Checking that there's exactly one input link
    if not self.inputs:
      raise IOError("The AutoDrive block should have an input link !")
    elif len(self.inputs) > 1:
      raise IOError("The AUtoDrive block can only have one input link !")

    self._link = self.inputs[0]

    # Opening and initializing the actuator to drive
    actuator_name = self._actuator.pop('name')
    self._device = actuator_list[actuator_name](**self._actuator)
    self._device.open()
    self._device.set_speed(0)

  def loop(self) -> None:
    """Receives the latest data from the VideoExtenso block, calculates the
    center coordinate in the chosen direction, and sets the actuator speed
    accordingly."""

    # Receiving the latest data
    data = self._link.recv_last(blocking=False)
    if data is None:
      return

    # Extracting the coordinates of the spots
    coord = data['Coord(px)']
    t = time()

    # Getting the average coordinate in the chosen direction
    y, x = list(zip(*coord))
    if 'x' in self._direction.lower():
      center = (max(x) + min(x)) / 2
    else:
      center = (max(y) + min(y)) / 2

    # Calculating the new speed to set
    diff = center - self._pixel_range / 2
    speed = max(-self._max_speed, min(self._max_speed, self._gain * diff))

    # Setting the speed and sending to downstream blocks
    self._device.set_speed(speed)
    self.send([t - self.t0, diff])

  def finish(self) -> None:
    """Simply sets the device speed to `0`."""

    self._device.set_speed(0)



# coding: utf-8

"""
Very simple program that displays the output of a chosen camera.

Required hardware:
  - Any camera (select the desired camera in the terminal)
"""

import crappy

if __name__ == "__main__":
  cam_list = list(crappy.camera.MetaCam.classes.keys())
  for i, c in enumerate(cam_list):
    print(i, c)
  r = int(input("What cam do you want to use ?> "))
  cam = cam_list[r]
  camera = crappy.blocks.Camera(camera=cam, verbose=True, display_images=True,
                                displayer_framerate=20)

  crappy.start()



# coding: utf-8

"""
Quick example showing how the PID block can be used to control the speed of a
motor.

The motor is a virtual entity, taking a voltage and returning a speed and
position.

No hardware required.
"""

import crappy

if __name__ == "__main__":

  g = crappy.blocks.Generator([
    {'type': 'constant', 'value': 1000, 'condition': 'delay=3'},
    {'type': 'ramp', 'speed': 100, 'condition': 'delay=5', 'init_value': 0},
    {'type': 'constant', 'value': 1800, 'condition': 'delay=3'},
    {'type': 'constant', 'value': 500, 'condition': 'delay=3'},
    {'type': 'sine', 'amplitude': 2000, 'offset': 1000, 'freq': .3,
     'condition': 'delay=15'}
  ], spam=True)

  kv = 1000

  mot = crappy.blocks.Machine([{'type': 'Fake_motor',
                                'cmd': 'pid',
                                'mode': 'speed',
                                'speed_label': 'speed',
                                # Motor properties:
                                'kv': kv,
                                'inertia': 4,
                                'rv': .2,
                                'fv': 1e-5
                                }])
  graph_m = crappy.blocks.Grapher(('t(s)', 'speed'), ('t(s)', 'cmd'))
  # , interp=False)

  crappy.link(mot, graph_m)
  crappy.link(g, graph_m)
  # To see what happens without PID
  # crappy.link(g, mot)
  # crappy.start()

  p = 38 / kv
  i = 2
  d = .05

  pid = crappy.blocks.PID(kp=p,
                          ki=i,
                          kd=d,
                          out_max=10,
                          out_min=-10,
                          i_limit=(-5, 5),
                          input_label='speed',
                          send_terms=True)

  crappy.link(g, pid)
  crappy.link(pid, mot)
  # crappy.link(mot, pid)  # This line will not smooth the feedback
  crappy.link(mot, pid, modifier=crappy.modifier.Moving_avg(15))

  graph_pid = crappy.blocks.Grapher(('t(s)', 'pid'))
  crappy.link(pid, graph_pid)

  graph_pid2 = crappy.blocks.Grapher(('t(s)', 'p_term'),
                                     ('t(s)', 'i_term'),
                                     ('t(s)', 'd_term'))

  crappy.link(pid, graph_pid2)

  crappy.start()



# coding: utf-8

from time import time, strftime, gmtime
from typing import Tuple, Optional, Dict, Any
import numpy as np
from .camera import Camera


class Fake_camera(Camera):
  """This camera class generates images without requiring any actual camera or
  existing image file.

  The generated images are just a gradient of grey levels, with a line moving
  as a function of time. It is possible to tune the dimension of the image, the
  frame rate and the speed of the line.
  """

  def __init__(self) -> None:
    """Initializes the parent class and instantiates the settings."""

    super().__init__()
    self._frame_nr = -1

    self.add_scale_setting('width', 1, 4096, None, self._gen_image, 1280)
    self.add_scale_setting('height', 1, 4096, None, self._gen_image, 1024)
    self.add_scale_setting('speed', 0., 800., None, None, 100.)
    self.add_scale_setting('fps', 0.1, 100., None, None, 50.)

  def open(self, **kwargs) -> None:
    """Sets the settings, generates the first image and initializes the time
    counter."""

    self.set_all(**kwargs)

    self._gen_image()

    self._t0 = time()
    self._t = self._t0

  def get_image(self) -> Tuple[Dict[str, Any], np.ndarray]:
    """Returns the updated image, depending only on the current timestamp.

    Also includes a waiting loop in order to achieve the right frame rate.
    """

    # Waiting in order to achieve the right frame rate
    while time() - self._t < 1 / self.fps:
      pass

    self._t = time()
    self._frame_nr += 1

    # Splitting the image to make a moving line
    row = int(self.speed * (self._t - self._t0)) % self.height
    metadata = {'t(s)': self._t,
                'DateTimeOriginal': strftime("%Y:%m:%d %H:%M:%S",
                                             gmtime(self._t)),
                'SubsecTimeOriginal': f'{self._t % 1:.6f}',
                'ImageUniqueID': self._frame_nr}
    return metadata, np.concatenate((self._img[row:], self._img[:row]), axis=0)

  def _gen_image(self, _: Optional[float] = None) -> None:
    """Generates the base gradient image, that will be splitted and returned
    in the :meth:`get_image` method"""

    self._img = np.arange(self.height) * 255. / self.height
    self._img = np.repeat(self._img.reshape(self.height, 1),
                          self.width, axis=1).astype(np.uint8)



# coding: utf-8

from typing import List, Optional
from .._global import OptionalModule
from .cameraConfigTools import Box
from .fields import get_res, get_field, allowed_fields

try:
  import cv2
except (ModuleNotFoundError, ImportError):
  cv2 = OptionalModule("opencv-python")
import numpy as np


class DISCorrel:
  """This class simply stores the data needed for performing image correlation
  using DisFlow, and also performs the correlation.

  Multiple parameters of DisFlow can be tuned, but not all of them. The class
  also takes care of projecting the image on the chosen base of fields, and
  optionally to estimate the residuals of the correlation process.
  """

  def __init__(self,
               fields: Optional[List[str]] = None,
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               init: bool = True,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3) -> None:
    """Sets the parameters of DisFlow.

    Args:
      fields: The base of fields to use for the projection, given as a
        :obj:`list` of :obj:`str`. The available fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

      alpha: Weight of the smoothness term in DisFlow.
      delta: Weight of the color constancy term in DisFlow.
      gamma: Weight of the gradient constancy term in DisFlow
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DisFlow (`0` means full scale).
      init: If :obj:`True`, the last field is used to initialize the
        calculation for the next one.
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DisFlow.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DisFlow.
      patch_size: Size of an image patch for matching in DisFlow
        (in pixels).
      patch_stride: Stride between neighbor patches in DisFlow. Must be
        less than patch size.
    """

    if fields is not None and not all((field in allowed_fields
                                       for field in fields)):
      raise ValueError(f"The only allowed values for the fields "
                       f"are {allowed_fields}")
    self._fields = ["x", "y", "exx", "eyy"] if fields is None else fields
    self._init = init

    # These attributes will be set later
    self._img0 = None
    self._height, self._width = None, None
    self._box = None
    self._dis_flow = None
    self._base = None
    self._norm2 = None

    # Setting the parameters of Disflow
    self._dis = cv2.DISOpticalFlow_create(cv2.DISOPTICAL_FLOW_PRESET_FAST)
    self._dis.setVariationalRefinementAlpha(alpha)
    self._dis.setVariationalRefinementDelta(delta)
    self._dis.setVariationalRefinementGamma(gamma)
    self._dis.setFinestScale(finest_scale)
    self._dis.setVariationalRefinementIterations(iterations)
    self._dis.setGradientDescentIterations(gradient_iterations)
    self._dis.setPatchSize(patch_size)
    self._dis.setPatchStride(patch_stride)

  def set_img0(self, img0: np.ndarray) -> None:
    """Sets the initial image to use for the correlation."""

    self._img0 = img0
    self._height, self._width, *_ = img0.shape
    self._dis_flow = np.zeros((self._height, self._width, 2))

  def set_box(self, box: Box) -> None:
    """Sets the region of interest to use for the correlation, and initializes
    other attributes."""

    # Sets the bounding box
    self._box = box
    x_top, x_bottom, y_left, y_right = box.sorted()
    box_height = y_right - y_left
    box_width = x_bottom - x_top

    # Creates and populates the base fields to use for correlation
    fields = np.empty((box_height, box_width, 2, len(self._fields)),
                      dtype=np.float32)
    for i, string in enumerate(self._fields):
      fields[:, :, 0, i], fields[:, :, 1, i] = get_field(string, box_height,
                                                         box_width)

    # These attributes will be used later
    self._base = [fields[:, :, :, i] for i in range(fields.shape[3])]
    self._norm2 = [np.sum(base_field ** 2) for base_field in self._base]

  def get_data(self,
               img: np.ndarray,
               residuals: bool = False) -> List[float]:
    """Processes the input image and returns the required data in a
    :obj:`list`.

    Args:
      img: The new image to process.
      residuals: Whether the residuals should be calculated or not for the
        image.

    Returns:
      A :obj:`list` containing the data to calculate, and the residuals at the
      end if requested.
    """

    # Making sure the reference image and the base fields were set
    if self._img0 is None:
      raise ValueError("The method set_img0 must be called first for setting "
                       "the reference image !")
    elif self._base is None:
      raise ValueError("The method set_box must be called first for setting "
                       "the region of interest !")

    # Updating the optical flow with the latest image
    if self._init:
      self._dis_flow = self._dis.calc(self._img0, img, self._dis_flow)
    else:
      self._dis_flow = self._dis.calc(self._img0, img, None)

    # Getting the values to calculate as floats
    ret = [np.sum(vec * self._crop(self._dis_flow)) / n2 for vec, n2 in
           zip(self._base, self._norm2)]

    # Adding the average residual value if requested
    if residuals:
      ret.append(np.average(np.abs(get_res(self._img0, img, self._dis_flow))))

    return ret

  def _crop(self, img: np.ndarray) -> np.ndarray:
    """Crops the image to the given region of interest."""

    x_min, x_max, y_min, y_max = self._box.sorted()
    return img[y_min:y_max, x_min:x_max]



# coding: utf-8

"""
Demonstration of how to create a custom Block in Crappy.

This block is intended to be used as a template, it doesn't actually act on
any device nor do anything exceptional.

No hardware required.
"""

import crappy
from time import time

# This class can be used as a starting point to create a new block. All methods
# except loop are optional.
# To add it to crappy, make the imports relative (refer to any other block),
# move the class to a file in crappy/blocks and add the corresponding line
# in crappy/blocks/__init__.py


# crappy.Block is a shortcut for crappy.blocks.Block
class MyBlock(crappy.Block):
  """A basic example of block.

  It will simply send a sawtooth signal.
  """

  def __init__(self, period=1):  # Optional
    # If you define your own constructor, do not forget to init Block !
    super().__init__()
    # Example argument, here the period of the sawtooth
    self.period = period

    # These attributes can be specified, they will be interpreted by Block
    self.freq = 100  # Hz
    self.labels = ['t(s)', 'cmd(V)']

  def prepare(self):  # Optional
    # Called inside the Process, before the actual beginning of the test
    # Create the objects needed for the test here
    # (ie. open a camera, a serial port etc...)
    print("==> prepare has been called")

  def begin(self):  # Optional
    # Now the test has begun, thi method will be called once
    print("==> begin has been called")
    # Send -1 only on the first loop
    # self.t0 contains the timestamp of the beginning of the test
    self.send([time() - self.t0, -1])

  def loop(self):
    # This loop will be called continuously until the end of the test
    # print("==> loop")
    t = time() - self.t0
    # Because self.labels is defined, we can send a list of the same
    # length, each value will be assigned to a label
    self.send([t, t % self.period])
    # But we could have written:
    # self.send({'t(s)': t, 'cmd(V)': t % self.period})

  def finish(self):  # Optional
    # This will be called before ending the process
    print("==> finish has been called")


if __name__ == "__main__":
  mb = MyBlock()
  # NB: matplotlib must be installed to used the Grapher
  graph = crappy.blocks.Grapher(('t(s)', 'cmd(V)'))
  crappy.link(mb, graph)

  crappy.start()



# coding: utf-8

from typing import Union, Dict
from .path import Path, condition_type


class Constant(Path):
  """The simplest path, simply sends the same value until the condition is
  met."""

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               condition: Union[str, condition_type],
               value: float = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      condition: The condition for switching to the next path. Refer to
        :ref:`generator path` for more info.
      value: The value to send.
    """

    Path.__init__(self, _last_time, _last_cmd)

    self._condition = self.parse_condition(condition)

    if value is None and _last_cmd is None:
      raise ValueError('For the first path, a value must be given !')

    self._value = _last_cmd if value is None else value

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met."""

    # Checking if the stop condition is met
    if self._condition(data):
      raise StopIteration

    # Returning the value
    return self._value



# coding: utf-8

"""
Demonstration of how to create a custom Actuator in Crappy.

This actuator is intended to be used as a template, it doesn't actually act on
any device.

No hardware required.
"""

import crappy
from time import sleep

# This class can be used as a starting point to create a new Actuator object
# To add it to crappy, make the imports relative (refer to any other
# actuators),
# move the class to a file in crappy/actuator and add the corresponding line
# in crappy/actuator/__init__.py


class My_actuator(crappy.actuator.Actuator):
  """A basic example of an Actuator object."""

  def __init__(self):
    # Do not forget to init Actuator !
    super().__init__()

  def open(self):
    print("Opening device...")
    sleep(1)
    self.pos = 0
    print("Device opened!")

  def close(self):
    print("Closing device...")
    sleep(.5)
    print("Device closed")

  # At least one of the two following methods (set_speed and set_position)
  # needs to be defined. Getters (get_speed and get_position) can be defined
  # too if the actuator supports it.
  def set_position(self, target, _=None):
    self.pos = target

  def get_position(self):
    return self.pos + 0.1  # To illustrate the difference with the target

  def stop(self):
    """Called before closing, should stop the actuator."""

    print("Stopping the actuator")


if __name__ == '__main__':
  sine_path = {'type': 'sine',
               'amplitude': 1,
               'freq': 1,
               'condition': None}

  gen = crappy.blocks.Generator([sine_path],
                                cmd_label='target_position',)

  machine = crappy.blocks.Machine([{
      'type': 'My_actuator',  # The class to instantiate
      'mode': 'position',  # set_position will be called
      'cmd': 'target_position',  # This actuator will be driven using the sine
      'pos_label': 'measured_position'}])  # The label to send the measurements

  crappy.link(gen, machine)

  graph = crappy.blocks.Grapher(('t(s)', 'measured_position'),
                                ('t(s)', 'target_position'))

  crappy.link(machine, graph)
  crappy.link(gen, graph)

  crappy.start()



# coding: utf-8

import crappy

help_string = """
Syntax:
{axe}{speed}
(You can put a space between axe and speed)
Example:
  x -5
Will pull the axe in at 5mm/s
  a20
Will push out motor A only (the number 1)
note: on this machine, motors are numbered from 1 to 4. Here they are labelled
a,b,c,d. (a is 1, b is 2,...)

Use q to quit
Type 'h' or 'help' to see this message again
"""


actuator_list = list()
actuator_list.append(crappy.actuator.Biaxe(port='/dev/ttyS4'))  # Motor 1
actuator_list.append(crappy.actuator.Biaxe(port='/dev/ttyS5'))  # Motor 2
actuator_list.append(crappy.actuator.Biaxe(port='/dev/ttyS6'))  # Motor 3
actuator_list.append(crappy.actuator.Biaxe(port='/dev/ttyS7'))  # Motor 4
for act in actuator_list:
  act.open()

actuator_dict = {}
i = 0
for c in 'abcd':
  actuator_dict[c] = [actuator_list[i]]
  i += 1
actuator_dict['x'] = actuator_dict['c']+actuator_dict['d']
actuator_dict['y'] = actuator_dict['a']+actuator_dict['b']

print(help_string)

user_input = ''
while True:
  user_input = input("Axe and speed ?> ").lower()
  if user_input == 'q':
    break
  elif user_input in ['h', 'help', '?']:
    print(help_string)
    continue
  try:
    axe = user_input[:1]
    speed = float(user_input[1:].strip())
  except ValueError:
    print("Unknown command, stopping")
    for axe in actuator_list:
      axe.set_speed(0)
    continue

  if user_input[0] in 'xyabcd':
    for axe in actuator_dict[user_input[0]]:
      axe.set_speed(speed)



# coding: utf-8

import numpy as np
from typing import Optional, Tuple
from _tkinter import TclError

from .block import Block
from .._global import OptionalModule

try:
  import matplotlib.pyplot as plt
  from matplotlib.widgets import Button
except (ModuleNotFoundError, ImportError):
  plt = OptionalModule("matplotlib")
  Button = OptionalModule("matplotlib")


class Grapher(Block):
  """The grapher receive data from a block and plots it.

  Multiple curves can be plotted on a same graph, and the data can come from
  different blocks.

  Note:
    To reduce the memory and CPU usage of graphs, try lowering the ``maxpt``
    parameter (2-3000 is already enough to follow a short test), or set the
    ``length`` parameter to a non-zero value (again, 2-3000 is fine). Lowering
    the ``freq`` is also a good option to limit the CPU use.
  """

  def __init__(self,
               *labels: Tuple[str, str],
               length: int = 0,
               freq: float = 2,
               maxpt: int = 20000,
               window_size: Tuple[int, int] = (8, 8),
               window_pos: Optional[Tuple[int, int]] = None,
               interp: bool = True,
               backend: str = "TkAgg",
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      *labels (:obj:`tuple`): Each :obj:`tuple` corresponds to a curve to plot,
        and should contain two values: the first will be the label of the `x`
        values, the second the label of the `y` values. There's no limit to the
        number of curves. Note that all the curves are displayed in a same
        graph.
      length (:obj:`int`, optional): If `0` the graph is static and displays
        all data from the start of the assay. Else only displays the last
        ``length`` received chunks, and drops the previous ones.
      freq (:obj:`float`, optional): The refresh rate of the graph. May cause
        high CPU use if set too high.
      maxpt (:obj:`int`, optional): The maximum number of points displayed on
        the graph. When reaching this limit, the block deletes one point out of
        two to avoid using too much memory and CPU.
      window_size (:obj:`tuple`, optional): The size of the graph, in inches.
      window_pos (:obj:`tuple`, optional): The position of the graph in pixels.
        The first value is for the `x` direction, the second for the `y`
        direction. The origin is the top-left corner. Works with multiple
        screens.
      interp (:obj:`bool`, optional): If :obj:`True`, the data points are
        linked together by straight lines. Else, only the points are displayed.
      backend (:obj:`int`, optional): The :mod:`matplotlib` backend to use.
        Performance may vary according to the chosen backend. Also, every
        backend may not be available depending on your machine.
      verbose (:obj:`bool`, optional): To display the loop frequency of the
        block.

    Example:
      ::

        graph = Grapher(('t(s)', 'F(N)'), ('t(s)', 'def(%)'))

      will plot a dynamic graph with two lines plot (`F=f(t)` and `def=f(t)`).
      ::

        graph = Grapher(('def(%)', 'F(N)'), length=0)

      will plot a static graph.
      ::

        graph = Grapher(('t(s)', 'F(N)'), length=30)

      will plot a dynamic graph displaying the last 30 chunks of data.
    """

    Block.__init__(self)
    self.niceness = 10
    self._length = length
    self.freq = freq
    self._maxpt = maxpt
    self._window_size = window_size
    self._window_pos = window_pos
    self._interp = interp
    self._backend = backend
    self.verbose = verbose

    self._labels = labels

  def prepare(self) -> None:

    # Switch to the required backend
    if self._backend:
      plt.switch_backend(self._backend)

    # Create the figure and the subplot
    self._figure = plt.figure(figsize=self._window_size)
    self._canvas = self._figure.canvas
    self._ax = self._figure.add_subplot(111)

    # Add the lines or the dots
    self._lines = []
    for _ in self._labels:
      if self._interp:
        self._lines.append(self._ax.plot([], [])[0])
      else:
        self._lines.append(self._ax.plot([], [], 'o', markersize=3)[0])

    # Keep only 1/factor points on each line
    self._factor = [1 for _ in self._labels]
    # Count to drop exactly 1/factor points, no more and no less
    self._counter = [0 for _ in self._labels]

    # Add the legend
    legend = [y for x, y in self._labels]
    plt.legend(legend, bbox_to_anchor=(-0.03, 1.02, 1.06, .102), loc=3,
               ncol=len(legend), mode="expand", borderaxespad=1)
    plt.xlabel(self._labels[0][0])
    plt.ylabel(self._labels[0][1])

    # Add a grid
    plt.grid()

    # Adds a button for clearing the graph
    self._clear_button = Button(plt.axes([.8, .02, .15, .05]), 'Clear')
    self._clear_button.on_clicked(self._clear)

    # Set the dimensions if required
    if self._window_pos:
      mng = plt.get_current_fig_manager()
      mng.window.wm_geometry("+%s+%s" % self._window_pos)

    # Ready to show the window
    plt.show(block=False)
    plt.pause(.001)

  def loop(self) -> None:

    # Receives the data sent by the upstream blocks
    if self.freq >= 10:
      # Assuming that above 10Hz the data won't saturate the links
      data = self.recv_all_delay()
    else:
      # Below 10Hz, making sure to flush the pipes at least every 0.1s
      data = self.recv_all_delay(delay=1 / 2 / self.freq,
                                 poll_delay=min(0.1, 1 / 2 / self.freq))

    update = False  # Should the graph be updated ?

    # For each curve, looking for the corresponding labels in the received data
    for i, (lx, ly) in enumerate(self._labels):
      x, y = None, None
      for dict_ in data:
        if lx in dict_ and ly in dict_:
          # Found the corresponding data, getting the new data according to the
          # current resampling factors
          dx = dict_[lx][self._factor[i] - self._counter[i] -
                         1::self._factor[i]]
          dy = dict_[ly][self._factor[i] - self._counter[i] -
                         1::self._factor[i]]
          # Recalculating the counter
          self._counter[i] = (self._counter[i] +
                              len(dict_[lx])) % self._factor[i]
          # Adding the new points to the arrays
          x = np.hstack((self._lines[i].get_xdata(), dx))
          y = np.hstack((self._lines[i].get_ydata(), dy))
          # the graph will need to be updated
          update = True
          # As we found the data, no need to search any further
          break

      # In case no matching labels were found, aborting for this curve
      if x is None:
        continue

      # Adjusting the number of points to remain below the length limit
      if self._length and len(x) >= self._length:
        x = x[-self._length:]
        y = y[-self._length:]

      # Dividing the number of points by two to remain below the maxpt limit
      elif len(x) > self._maxpt:
        print(f"[Grapher] Too many points on the graph "
              f"{i} ({len(x)}>{self._maxpt})")
        x, y = x[::2], y[::2]
        self._factor[i] *= 2
        print(f"[Grapher] Resampling factor is now {self._factor[i]}")

      # Finally, updating the data on the graph
      self._lines[i].set_xdata(x)
      self._lines[i].set_ydata(y)

    # Updating the graph if necessary
    if update:
      self._ax.relim()
      self._ax.autoscale()
      try:
        self._canvas.draw()
      except TclError:
        pass
      self._canvas.flush_events()

  def finish(self) -> None:
    plt.close("all")

  def _clear(self, *_, **__) -> None:
    for line in self._lines:
      line.set_xdata([])
      line.set_ydata([])
    self.factor = [1 for _ in self._labels]
    self.counter = [0 for _ in self._labels]



# coding: utf-8

"""
Demonstration of how to create a custom Camera in Crappy.

This camera is intended to be used as a template, it doesn't actually display
any image from a real camera.

No hardware required.
"""

import crappy
import numpy as np
from time import time

# This class can be used as a starting point to create a new Camera.
# To add it to crappy, make the imports relative (refer to any other camera),
# move the class to a file in crappy/camera and add the corresponding line
# in crappy/camera/__init__.py


class My_cam(crappy.camera.Camera):
  """A basic example of Camera.

  It will simply send a gray image, the level can be set using a
  Camera_setting.
  """

  def __init__(self, resolution=(480, 640)):
    # Do not forget to init Camera !
    super().__init__()
    self._resolution = resolution
    self._frame = None

    # Optional: add settings to the camera
    # Here we will set the gray level of the image
    self.add_scale_setting('level', 0, 255, self._get_lvl, self._set_lvl, 128)

  def _get_lvl(self):
    """The gray level getter."""

    return self._frame[0, 0]

  def _set_lvl(self, val):
    """The gray level setter.

    Recreates the image with the new level.
    """

    self._frame = np.ones(self._resolution, dtype=np.uint8) * val
    return val

  def open(self, **kwargs):
    """Will be called in the .prepare() of the block."""

    # Let's create our image
    self._frame = np.zeros(self._resolution, dtype=np.uint8)
    # Allow delegation of generic camera args such as max_fps
    self.set_all(**kwargs)

  def get_image(self):
    """The method that returns the frame. It must also return the time."""

    return time(), self._frame

  def close(self):
    """Will be called on exit or crash."""

    del self._frame


if __name__ == '__main__':
  cam = crappy.blocks.Camera('My_cam',
                             freq=60,
                             display_images=True,
                             verbose=True)

  crappy.start()
